"""
A HTTP fetcher class.
"""

import os
import urlparse
import requests
import shutil
import mimetypes
from teapot.extra.rfc6266 import parse_requests_response

from teapot.log import LOGGER
from teapot.fetchers.base_fetcher import BaseFetcher


class HttpFetcher(BaseFetcher):

    """
    Fetchs a file from the web, using HTTP or HTTPS.
    """

    shortname = 'http'

    def read_source(self, source):
        """
        Checks that the `source` is an HTTP(S) URL.
        """

        assert(source)

        url = urlparse.urlparse(source.location)

        if url.scheme in ['http', 'https']:
            self.url = urlparse.urlunparse(url)

            return True

    def do_fetch(self, target):
        """
        Fetch the archive at the specified URL.
        """

        response = requests.get(self.url, stream=True)
        response.raise_for_status()

        mimetype = response.headers.get('content-type')
        encoding = response.headers.get('content-encoding')
        archive_type = (mimetype, encoding)

        # If the source has an overriden type, we use that instead.
        extension = None

        if self.source._type:
            extension = mimetypes.guess_extension(self.source._type[0])

        if not extension:
            extension = mimetypes.guess_extension(mimetype)

        if not extension:
            LOGGER.debug('No extension registered for this mimetype (%s). Guessing one from the URL...', mimetype)

            extension = os.path.splitext(urlparse.urlparse(self.url).path)[1]

        if extension and extension.startswith('.'):
            extension = extension[1:]

        content_disposition = parse_requests_response(response)

        filename = content_disposition.filename_sanitized(extension=extension, default_filename='archive')

        content_length = response.headers.get('content-length')

        if content_length is not None:
            content_length = int(content_length)

        archive_path = os.path.join(target, filename)

        self.progress.on_start(target=os.path.basename(archive_path), size=content_length)

        with open(archive_path, 'wb') as target_file:
            current_size = 0

            for buf in response.iter_content(1024):

                if buf:
                    target_file.write(buf)
                    current_size += len(buf)

                    self.progress.on_update(progress=current_size)

        self.progress.on_finish()

        return {
            'archive_path': archive_path,
            'archive_type': archive_type,
        }



"""
The filters.
"""

from teapot.filters.builtin import *
from teapot.filters.decorators import get_filter_by_name


class Filtered(object):
    """
    Inherit from this class when your object can be disabled by filters.
    """

    def __init__(self, filters=[]):
        """
        Set the `filters` to check.

        `filters` can be None, a string, or a list of strings that will be
        resolved.
        """

        if not filters:
            self.filters = []
        elif isinstance(filters, basestring):
            self.filters = [get_filter_by_name(filters)]
        else:
            self.filters = map(get_filter_by_name, filters)

    @property
    def enabled(self):
        """
        Check if the instance is filtered out.
        """

        return all(f() for f in self.filters)



"""
A base Unpacker class.
"""

# TODO: Implement this
#from teapot.unpacker.callbacks import NullUnpackerCallback


class DuplicateUnpackerMimetypeError(ValueError):

    """
    An already existing mimetype was specified for a new BaseUnpacker subclass.
    """

    def __init__(self, mimetype, encoding, original_class, new_class):
        """
        Create a DuplicateUnpackerMimetypeError for the specified `mimetype`,
        `encoding`, `original_class` and `new_class`.
        """

        super(DuplicateUnpackerMimetypeError, self).__init__(
            'Cannot assign mimetype/encoding %r to %s as is already registered to %s' % (
                (mimetype, encoding),
                original_class,
                new_class,
            )
        )


class BaseUnpacker(object):

    """
    Base class for all unpacker classes.

    If you subclass this class, you will have to re-implement the
    :func:`do_unpack` method to provide your specific unpacker logic.

    If you desire to attach your unpacker to certain mimetypes, please do so by
    defining the class-level member `types` that must be a list of couples
    `(mimetype, encoding)`.
    """

    index = {}

    class __metaclass__(type):

        """
        A metaclass that automatically registers all BaseUnpacker subclasses
        that provide mimetype and encoding attributes in their definition.
        """

        def __init__(cls, name, bases, attrs):
            """
            Registers the class into the index if it has a shortname attribute
            in its definition.
            """

            type.__init__(cls, name, bases, attrs)

            if 'types' in attrs and cls.types:
                if not isinstance(cls.types, list):
                    raise TypeError(
                        'The types attribute for %s must be a list of 2-tuples.' %
                        cls)

                for mimetype, encoding in cls.types:
                    if (mimetype, encoding) in BaseUnpacker.index:
                        raise DuplicateUnpackerMimetypeError(
                            mimetype=mimetype,
                            encoding=encoding,
                            original_class=BaseUnpacker.index[(mimetype, encoding)],
                            new_class=cls,
                        )
                    else:
                        BaseUnpacker.index[(mimetype, encoding)] = cls
            else:
                # We ensure all unpacker classes have at least an empty
                # types attribute.
                setattr(cls, 'types', None)

    def __init__(self, attendee):
        """
        Create a new unpacker.
        """

        if not attendee:
            raise ValueError('An unpacker must be associated to an attendee.')

        self.attendee = attendee
        self.progress = self.attendee.party.unpacker_callback_class(self)

    def __repr__(self):
        """
        Get the representation of the unpacker.
        """

        return '<%s.%s()>' % (
            self.__class__.__module__,
            self.__class__.__name__,
        )

    @property
    def archive_path(self):
        """
        Get the archive path.

        This is a shortcut function for `self.attendee.archive_path`.
        """

        return self.attendee.archive_path

    def unpack(self):
        """
        Unpack the associated attendee archive.
        """

        try:
            return self.do_unpack()

        except Exception as ex:
            self.progress.on_exception(ex)

            raise

    def do_unpack(self):
        """
        Unpack an archive.

        The archive to unpack can be reached at `self.archive_path`.

        This method must return a dict with the following keys:
            - `source_tree_path`: The extracted source tree path.

        It must raise an exception on error.

        You can provide feedback on the unpacking operation by calling
        `self.progress.on_start`, `self.progress.on_update` and
        `self.progress.on_finish` at the appropriate time.

        See :class:`teapot.unpackers.callbacks.BaseUnpackerCallback` for
        further details.
        """

        raise NotImplementedError



"""
teapot 'Source' class.
"""

import os
import json
import hashlib

from teapot.log import LOGGER
from teapot.log import Highlight as hl
from teapot.fetchers import get_fetcher_class_from_shortname, guess_fetcher_instance
from teapot.fetchers.callbacks import NullFetcherCallback
from teapot.filters import Filtered


def read_type(_type):
    """
    Read a type in different formats.

    If `_type` is a falsy value, nothing is returned.

    If `_type` is a string, a 2-tuple (`_type`, None) will be returned.

    Otherwise `_type` will be assumed to be an iterable. Its size must not
    exceed 2 elements, and it will be returned as a 2-tuple.
    """

    if not _type:
        return

    if isinstance(_type, basestring):
        return (_type, None)

    if len(_type) == 1:
        return tuple(_type[0], None)
    elif len(_type) == 2:
        return tuple(_type)

    raise ValueError('Incorrect type: %r' % _type)

def make_sources(attendee, sources):
    """
    Build a list of Source instances.

    `sources` can be either a simple location string, a dictionary, or an
    iterable of strings and/or dictionaries.

    If `sources` is false, an empty list is returned.
    """

    if not sources:
        return []

    elif isinstance(sources, basestring):
        return [
            Source(
                attendee=attendee,
                location=unicode(sources),
                _type=None,
                fetcher_class=guess_fetcher_instance,
                fetcher_options={},
            ),
        ]

    elif isinstance(sources, dict):
        return [
            Source(
                attendee=attendee,
                location=sources.get('location'),
                _type=read_type(sources.get('type')),
                fetcher_class=get_fetcher_class_from_shortname(
                    sources.get('fetcher')
                ),
                fetcher_options=sources.get('fetcher_options'),
                filters=sources.get('filters'),
            ),
        ]

    return sum(map(lambda x: make_sources(attendee, x), sources), [])


class Source(Filtered):

    """
    A Source instance holds information about where and how to get a
    third-party software.
    """

    def __init__(self, attendee, location, _type, fetcher_class, fetcher_options, filters=[]):
        """
        Create a Source instance.

        `attendee` is the attendee this source is attached to.

        `location` is the origin of the third-party software archive to get.
        Its format an meaning depends on the associated `fetcher_class`.

        `_type` is a 2-tuple containing the mimetype, and the encoding of the
        source. If `_type` is falsy, the type will be guessed from the source
        itself.

        `fetcher_options` is a free-format structure that will be passed as a
        parameter to the fetcher on instanciation.

        `filters` is a list of filters that must be truth for the source to be
        active.
        """

        if not attendee:
            raise ValueError('An source must be associated to an attendee.')

        self.attendee = attendee
        self.location = location
        self._type = _type
        self.fetcher_class = fetcher_class
        self.fetcher_options = fetcher_options
        self.__fetcher = None

        Filtered.__init__(self, filters=filters)

    def __repr__(self):
        """
        Get a representation of the source.
        """

        return '<%s.%s(location=%r, fetcher_class=%r)>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.location,
            self.fetcher_class,
        )

    def __str__(self):
        """
        Get a string representation of the source.
        """

        return self.location

    @property
    def fetcher(self):
        """
        Get the associated fetcher instance.
        """

        if self.__fetcher is None:
            self.__fetcher = self.fetcher_class(self)

        return self.__fetcher

    def fetch(self, root_path):
        """
        Fetch the specified source.

        Returns dict, containing the information about the archive.
        """

        if not self.fetcher:
            return

        try:
            cache_info = self.fetcher.fetch(
                target=root_path,
            )

            if self._type:
                cache_info['archive_type'] = self._type

            cache_info['archive_hash'] = self.hash_file(cache_info['archive_path'])
            cache_info['source_hash'] = self.source_hash

            return cache_info

        except Exception as ex:
            LOGGER.exception(ex)

    def hash_file(self, path):
        """
        Returns the hash of the specified file or directory.
        """

        if os.path.isdir(path):

            data = {}

            for root, dirs, files in os.walk(path):
                for _file in files:
                    data[os.path.join(root, _file)] = self.hash_file(os.path.join(root, _file))

            algorithm = hashlib.sha1()
            algorithm.update(json.dumps(data))

            result = algorithm.hexdigest()

        else:
            algorithm = hashlib.sha1()
            read_size = algorithm.block_size * 256

            with open(path, 'rb') as _file:
                data = _file.read(read_size)

                while data:
                    algorithm.update(_file.read())
                    data = _file.read(read_size)

            result = algorithm.hexdigest()

        return result

    @property
    def source_hash(self):
        """
        Get the source hash.
        """

        data = {
            'location': self.location,
            'type': self._type,
            'fetcher_class': self.fetcher_class.__name__,
            'fetcher_options': self.fetcher_options,
        }

        algorithm = hashlib.sha1()
        algorithm.update(json.dumps(data))

        result = algorithm.hexdigest()

        return result



"""
An unpacker that does nothing.
"""

from teapot.unpackers.base_unpacker import BaseUnpacker

import os
import shutil


class NullUnpacker(BaseUnpacker):

    """
    An unpacker class that just copy source trees.
    """

    types = [
        (None, None),
    ]

    def do_unpack(self):
        """
        Uncompress the archive.

        Return the path of the extracted folder.
        """

        if not os.path.isdir(self.archive_path):
            raise RuntimeError('A directory was expected.')

        source_tree_path = os.path.join(self.attendee.build_path, os.path.basename(self.archive_path))

        self.progress.on_start(count=1)

        self.progress.on_update(current_file=self.archive_path, progress=1)
        shutil.copytree(self.archive_path, source_tree_path)

        self.progress.on_finish()

        return {
            'source_tree_path': source_tree_path,
        }



# vim: set fileencoding=utf-8 sw=4 ts=4 et :

"""Implements RFC 6266, the Content-Disposition HTTP header.

parse_headers handles the receiver side.
It has shortcuts for some http libraries:
    parse_httplib2_response and parse_requests_response.
It returns a ContentDisposition object with attributes like is_inline,
filename_unsafe, filename_sanitized.

build_header handles the sender side.
"""

from lepl import *
from collections import namedtuple
from urllib import quote, unquote
from urlparse import urlsplit
from string import hexdigits, ascii_letters, digits

import posixpath
import os.path
import re
import sys

__all__ = (
    'ContentDisposition',
    'parse_headers',
    'parse_httplib2_response',
    'parse_requests_response',
    'build_header',
)


PY3K = sys.version_info >= (3,)

LangTagged = namedtuple('LangTagged', 'string langtag')


if PY3K:
    # XXX Both implementations allow stray %
    def percent_encode(string, safe, encoding):
        return quote(string, safe, encoding, errors='strict')

    def percent_decode(string, encoding):
        # unquote doesn't default to strict, fix that
        return unquote(string, encoding, errors='strict')
else:
    def percent_encode(string, **kwargs):
        encoding = kwargs.pop('encoding')
        return quote(string.encode(encoding), **kwargs)

    def percent_decode(string, **kwargs):
        encoding = kwargs.pop('encoding')
        return unquote(string, **kwargs).decode(encoding)


class ContentDisposition(object):
    """
    Records various indications and hints about content disposition.

    These can be used to know if a file should be downloaded or
    displayed directly, and to hint what filename it should have
    in the download case.
    """

    def __init__(self, disposition='inline', assocs=None, location=None):
        """This constructor is used internally after parsing the header.

        Instances should generally be created from a factory
        function, such as parse_headers and its variants.
        """

        self.disposition = disposition
        self.location = location
        if assocs is None:
            self.assocs = {}
        else:
            # XXX Check that parameters aren't repeated
            self.assocs = dict((key.lower(), val) for (key, val) in assocs)

    @property
    def filename_unsafe(self):
        """The filename from the Content-Disposition header.

        If a location was passed at instanciation, the basename
        from that may be used as a fallback. Otherwise, this may
        be the None value.

        On safety:
            This property records the intent of the sender.

            You shouldn't use this sender-controlled value as a filesystem
        path, it can be insecure. Serving files with this filename can be
        dangerous as well, due to a certain browser using the part after the
        dot for mime-sniffing.
        Saving it to a database is fine by itself though.
        """

        if 'filename*' in self.assocs:
            return self.assocs['filename*'].string
        elif 'filename' in self.assocs:
            # XXX Reject non-ascii (parsed via qdtext) here?
            return self.assocs['filename']
        elif self.location is not None:
            return posixpath.basename(self.location_path.rstrip('/'))

    @property
    def location_path(self):
        if self.location:
            return percent_decode(
                urlsplit(self.location, scheme='http').path,
                encoding='utf-8')

    def filename_sanitized(self, extension, default_filename='file'):
        """Returns a filename that is safer to use on the filesystem.

        The filename will not contain a slash (nor the path separator
        for the current platform, if different), it will not
        start with a dot, and it will have the expected extension.

        No guarantees that makes it "safe enough".
        No effort is made to remove special characters;
        using this value blindly might overwrite existing files, etc.
        """

        assert extension
        assert extension[0] != '.'
        assert default_filename
        assert '.' not in default_filename
        extension = '.' + extension

        fname = self.filename_unsafe
        if fname is None:
            fname = default_filename
        fname = posixpath.basename(fname)
        fname = os.path.basename(fname)
        fname = fname.lstrip('.')
        if not fname:
            fname = default_filename
        if not fname.endswith(extension):
            fname = fname + extension
        return fname

    @property
    def is_inline(self):
        """If this property is true, the file should be handled inline.

        Otherwise, and unless your application supports other dispositions
        than the standard inline and attachment, it should be handled
        as an attachment.
        """

        return self.disposition.lower() == 'inline'

    def __repr__(self):
        return 'ContentDisposition(%r, %r, %r)' % (
            self.disposition, self.assocs, self.location)


def ensure_charset(text, encoding):
    if isinstance(text, bytes):
        return text.decode(encoding)
    else:
        assert fits_inside_codec(text, encoding)
        return text


def parse_headers(content_disposition, location=None, relaxed=False):
    """Build a ContentDisposition from header values.
    """

    if content_disposition is None:
        return ContentDisposition(location=location)

    # Both alternatives seem valid.
    if False:
        # Require content_disposition to be ascii bytes (0-127),
        # or characters in the ascii range
        content_disposition = ensure_charset(content_disposition, 'ascii')
    else:
        # We allow non-ascii here (it will only be parsed inside of
        # qdtext, and rejected by the grammar if it appears in
        # other places), although parsing it can be ambiguous.
        # Parsing it ensures that a non-ambiguous filename* value
        # won't get dismissed because of an unrelated ambiguity
        # in the filename parameter. But it does mean we occasionally
        # give less-than-certain values for some legacy senders.
        content_disposition = ensure_charset(content_disposition, 'iso-8859-1')

    # Check the caller already did LWS-folding (normally done
    # when separating header names and values; RFC 2616 section 2.2
    # says it should be done before interpretation at any rate).
    # Hopefully space still means what it should in iso-8859-1.
    # This check is a bit stronger that LWS folding, it will
    # remove CR and LF even if they aren't part of a CRLF.
    # However http doesn't allow isolated CR and LF in headers outside
    # of LWS.

    if relaxed:
        # Relaxed has two effects (so far):
        # the grammar allows a final ';' in the header;
        # we do LWS-folding, and possibly normalise other broken
        # whitespace, instead of rejecting non-lws-safe text.
        # XXX Would prefer to accept only the quoted whitespace
        # case, rather than normalising everything.
        content_disposition = normalize_ws(content_disposition)
        parser = content_disposition_value_relaxed
    else:
        # Turns out this is occasionally broken: two spaces inside
        # a quoted_string's qdtext. Firefox and Chrome save the two spaces.
        if not is_lws_safe(content_disposition):
            raise ValueError(
                content_disposition, 'Contains nonstandard whitespace')

        parser = content_disposition_value

    try:
        parsed = parser.parse(content_disposition)
    except FullFirstMatchException:
        return ContentDisposition(location=location)
    return ContentDisposition(
        disposition=parsed[0], assocs=parsed[1:], location=location)


def parse_httplib2_response(response, **kwargs):
    """Build a ContentDisposition from an httplib2 response.
    """

    return parse_headers(
        response.get('content-disposition'),
        response['content-location'], **kwargs)


def parse_requests_response(response, **kwargs):
    """Build a ContentDisposition from a requests (PyPI) response.
    """

    return parse_headers(
        response.headers.get('content-disposition'), response.url, **kwargs)


def parse_ext_value(val):
    charset = val[0]
    if len(val) == 3:
        charset, langtag, coded = val
    else:
        charset, coded = val
        langtag = None
    if not PY3K and isinstance(coded, unicode):
        coded = coded.encode('ascii')
    decoded = percent_decode(coded, encoding=charset)
    return LangTagged(decoded, langtag)


# Currently LEPL doesn't handle case-insensivitity:
# https://groups.google.com/group/lepl/browse_thread/thread/68e7b136038772ca
def CaseInsensitiveLiteral(lit):
    return Regexp('(?i)' + re.escape(lit))


# RFC 2616
separator_chars = "()<>@,;:\\\"/[]?={} \t"
ctl_chars = ''.join(chr(i) for i in xrange(32)) + chr(127)
nontoken_chars = separator_chars + ctl_chars

# RFC 5987
attr_chars_nonalnum = '!#$&+-.^_`|~'
attr_chars = ascii_letters + digits + attr_chars_nonalnum

# RFC 5987 gives this alternative construction of the token character class
token_chars = attr_chars + "*'%"


# To debug, wrap in this block:
#with TraceVariables():

# Definitions from https://tools.ietf.org/html/rfc2616#section-2.2
# token was redefined from attr_chars to avoid using AnyBut,
# which might include non-ascii octets.
token = Any(token_chars)[1:, ...]


# RFC 2616 says some linear whitespace (LWS) is in fact allowed in text
# and qdtext; however it also mentions folding that whitespace into
# a single SP (which isn't in CTL) before interpretation.
# Assume the caller already that folding when parsing headers.

# NOTE: qdtext also allows non-ascii, which we choose to parse
# as ISO-8859-1; rejecting it entirely would also be permitted.
# Some broken browsers attempt encoding-sniffing, which is broken
# because the spec only allows iso, and because encoding-sniffing
# can mangle valid values.
# Everything else in this grammar (including RFC 5987 ext values)
# is in an ascii-safe encoding.
# Because of this, this is the only character class to use AnyBut,
# and all the others are defined with Any.
qdtext = AnyBut('"' + ctl_chars)

char = Any(''.join(chr(i) for i in xrange(128)))  # ascii range: 0-127

quoted_pair = Drop('\\') + char
quoted_string = Drop('"') & (quoted_pair | qdtext)[:, ...] & Drop('"')

value = token | quoted_string

# Other charsets are forbidden, the spec reserves them
# for future evolutions.
charset = (CaseInsensitiveLiteral('UTF-8')
           | CaseInsensitiveLiteral('ISO-8859-1'))

# XXX See RFC 5646 for the correct definition
language = token

attr_char = Any(attr_chars)
hexdig = Any(hexdigits)
pct_encoded = '%' + hexdig + hexdig
value_chars = (pct_encoded | attr_char)[...]
ext_value = (
    charset & Drop("'") & Optional(language) & Drop("'")
    & value_chars) > parse_ext_value
ext_token = token + '*'
noext_token = ~Lookahead(ext_token) & token

# Adapted from https://tools.ietf.org/html/rfc6266
# Mostly this was simplified to fold filename / filename*
# into the normal handling of ext_token / noext_token
with DroppedSpace():
    disposition_parm = (
        (ext_token & Drop('=') & ext_value)
        | (noext_token & Drop('=') & value)) > tuple
    disposition_type = (
        CaseInsensitiveLiteral('inline')
        | CaseInsensitiveLiteral('attachment')
        | token)
    content_disposition_value = (
        disposition_type & Star(Drop(';') & disposition_parm))

    # Allows nonconformant final semicolon
    # I've seen it in the wild, and browsers accept it
    # http://greenbytes.de/tech/tc2231/#attwithasciifilenamenqs
    content_disposition_value_relaxed = (
        content_disposition_value & Optional(Drop(';')))


def is_token_char(ch):
    # Must be ascii, and neither a control char nor a separator char
    asciicode = ord(ch)
    # < 128 means ascii, exclude control chars at 0-31 and 127,
    # exclude separator characters.
    return 31 < asciicode < 127 and ch not in separator_chars


def usesonlycharsfrom(candidate, chars):
    # Found that shortcut in urllib.quote
    return candidate.rstrip(chars) == ''


def is_token(candidate):
    #return usesonlycharsfrom(candidate, token_chars)
    return all(is_token_char(ch) for ch in candidate)


def is_ascii(text):
    return all(ord(ch) < 128 for ch in text)


def fits_inside_codec(text, codec):
    try:
        text.encode(codec)
    except UnicodeEncodeError:
        return False
    else:
        return True


def is_lws_safe(text):
    return normalize_ws(text) == text


def normalize_ws(text):
    return ' '.join(text.split())


def qd_quote(text):
    return text.replace('\\', '\\\\').replace('"', '\\"')


def build_header(
    filename, disposition='attachment', filename_compat=None
):
    """Generate a Content-Disposition header for a given filename.

    For legacy clients that don't understand the filename* parameter,
    a filename_compat value may be given.
    It should either be ascii-only (recommended) or iso-8859-1 only.
    In the later case it should be a character string
    (unicode in Python 2).

    Options for generating filename_compat (only useful for legacy clients):
    - ignore (will only send filename*);
    - strip accents using unicode's decomposing normalisations,
    which can be done from unicode data (stdlib), and keep only ascii;
    - use the ascii transliteration tables from Unidecode (PyPI);
    - use iso-8859-1
    Ignore is the safest, and can be used to trigger a fallback
    to the document location (which can be percent-encoded utf-8
    if you control the URLs).

    See https://tools.ietf.org/html/rfc6266#appendix-D
    """

    # While this method exists, it could also sanitize the filename
    # by rejecting slashes or other weirdness that might upset a receiver.

    if disposition != 'attachment':
        assert is_token(disposition)

    rv = disposition

    if is_token(filename):
        rv += '; filename=%s' % (filename, )
        return rv
    elif is_ascii(filename) and is_lws_safe(filename):
        qd_filename = qd_quote(filename)
        rv += '; filename="%s"' % (qd_filename, )
        if qd_filename == filename:
            # RFC 6266 claims some implementations are iffy on qdtext's
            # backslash-escaping, we'll include filename* in that case.
            return rv
    elif filename_compat:
        if is_token(filename_compat):
            rv += '; filename=%s' % (filename_compat, )
        else:
            assert is_lws_safe(filename_compat)
            rv += '; filename="%s"' % (qd_quote(filename_compat), )

    # alnum are already considered always-safe, but the rest isn't.
    # Python encodes ~ when it shouldn't, for example.
    rv += "; filename*=utf-8''%s" % (percent_encode(
        filename, safe=attr_chars_nonalnum, encoding='utf-8'), )

    # This will only encode filename_compat, if it used non-ascii iso-8859-1.
    return rv.encode('iso-8859-1')




"""
An unpacker that deals with .zip files.
"""

from teapot.unpackers.base_unpacker import BaseUnpacker

import os
import zipfile

class InvalidZipFileError(RuntimeError):

    """
    The specified archive is not a supported zip file.
    """

    def __init__(self, archive_path):
        """
        Create an InvalidZipFileError for the specified `archive_path`.
        """

        super(InvalidZipFileError, self).__init__(
            'Unsupported archive: %s' % archive_path
        )


class ZipFileHasNoCommonPrefixError(RuntimeError):

    """
    The specified archive does not contain a common prefix for its members.
    """

    def __init__(self, archive_path):
        """
        Create an ZipFileHasNoCommonPrefixError for the specified
        `archive_path`.
        """

        super(ZipFileHasNoCommonPrefixError, self).__init__(
            'The specified archive contains no common prefix: %s' % archive_path
        )


class ZipFileUnpacker(BaseUnpacker):

    """
    An unpacker class that deals with .zip files.
    """

    types = [
        ('application/zip', None),
    ]

    def do_unpack(self):
        """
        Uncompress the archive.

        Return the path of the extracted folder.
        """

        if not zipfile.is_zipfile(self.archive_path):
            raise InvalidZipFileError(archive_path=self.archive_path)

        zipf = zipfile.ZipFile(self.archive_path, 'r')

        # We get the common prefix for all archive members.
        prefix = os.path.commonprefix(zipf.namelist())

        while prefix and not prefix.endswith('/'):
            prefix = os.path.dirname(prefix)

        # We remove the trailing '/'
        prefix = os.path.dirname(prefix)

        if not prefix:
            raise ZipFileHasNoCommonPrefixError(archive_path=self.archive_path)

        source_tree_path = os.path.join(self.attendee.build_path, prefix)

        self.progress.on_start(count=len(zipf.infolist()))

        for index, info in enumerate(zipf.infolist()):
            if os.path.isabs(info.filename):
                raise ValueError('Refusing to extract archive that contains absolute filenames.')

            self.progress.on_update(current_file=info.filename, progress=index)
            zipf.extract(info, path=self.attendee.build_path)

        self.progress.on_finish()

        return {
            'source_tree_path': source_tree_path,
        }



"""
A base Fetcher class.
"""

from teapot.fetchers.callbacks import NullFetcherCallback


class DuplicateFetcherShortNameError(ValueError):

    """
    An already existing shortname was specified for a new BaseFetcher subclass.
    """

    def __init__(self, shortname, original_class, new_class):
        """
        Create a DuplicateFetcherShortNameError for the specified `shortname`,
        `original_class` and `new_class`.
        """

        super(DuplicateFetcherShortNameError, self).__init__(
            'Cannot assign shortname %r to %s as is already registered to %s' % (
                shortname,
                original_class,
                new_class,
            )
        )


class UnsupportedLocationError(ValueError):

    """
    A fetcher does not handle the specified location.
    """

    def __init__(self, source, fetcher_class):
        """
        Create an UnsupportedLocationError for the specified `source` and
        `fetcher_class`.
        """

        super(UnsupportedLocationError, self).__init__(
            'The fetcher class %s does not support the specified source (%s)' % (
                fetcher_class,
                source,
            )
        )


class BaseFetcher(object):

    """
    Base class for all fetcher classes.

    If you subclass this class, you will have to re-implement the
    :func:`read_source` and :func:`do_fetch` methods in your subclass to
    provide your fetcher specific fetch logic.

    If you intend to expose your fetcher to users, you must also declare a
    unique class-level `shortname`. This is what the user will specify in the
    :term:`party file` to use a specific fetcher.
    """

    index = {}

    class __metaclass__(type):

        """
        A metaclass that automatically registers all BaseFetcher subclasses
        that provide a shortname attribute in their definition.
        """

        def __init__(cls, name, bases, attrs):
            """
            Registers the class into the index if it has a shortname attribute
            in its definition.
            """

            type.__init__(cls, name, bases, attrs)

            if 'shortname' in attrs and cls.shortname:
                if not isinstance(cls.shortname, basestring):
                    raise TypeError(
                        'The shortname attribute for %s must be a string.' %
                        cls)

                if cls.shortname in BaseFetcher.index:
                    raise DuplicateFetcherShortNameError(
                        shortname=cls.shortname,
                        original_class=BaseFetcher.index[cls.shortname],
                        new_class=cls,
                    )
                else:
                    BaseFetcher.index[cls.shortname] = cls
            else:
                # We ensure all fetcher classes have at least an empty
                # shortname attribute.
                setattr(cls, 'shortname', None)

    def __init__(self, source):
        """
        Instantiate a base fetcher.

        `source` is the source.
        """

        if not source:
            raise ValueError('A fetcher must be associated to a source.')

        if not self.read_source(source):
            raise UnsupportedLocationError(
                source=source,
                fetcher_class=self.__class__,
            )

        self.source = source
        self.progress = source.attendee.party.fetcher_callback_class(self)

    def __repr__(self):
        """
        Get the representation of the fetcher.
        """

        return '<%s.%s(source=%r)>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.source,
        )

    def read_source(self, source):
        """
        Reimplement this method to indicate whether or not your fetcher class
        supports the specified `source`.

        This method is called during the fetcher creation and must return a
        truthy value if the fetcher supports the specified source.

        You may set custom instance variable in this method as it is guaranteed
        that it gets called before the :func:`do_fetch` method. For instance,
        one could parse a source URL and get the meaningful components into
        member variables for later use in :func:`do_fetch`.
        """

        raise NotImplementedError

    def fetch(self, target):
        """
        Fetch the associated source using `target` as a suggested filename.
        """

        try:
            return self.do_fetch(target)

        except Exception as ex:
            self.progress.on_exception(ex)

            raise

    def do_fetch(self, target):
        """
        Reimplement this method with your specific fetcher logic.

        `target` is the suggested target filename.

        This method must return a dict with the following keys:
            - `archive_path`: The archive absolute path.
            - `archive_type`: A couple containing the mimetype, then the
              encoding of the archive. Example: ``('application/x-gzip', None)``

        It must raise an exception on error.

        You can (and should) provide feedback on the fetching operation by
        calling `self.progress.on_start`, `self.progress.on_update` and
        `self.progress.on_finish` at the appropriate time.

        See :class:`teapot.fetchers.callbacks.BaseFetcherCallback`
        documentation for details.
        """

        raise NotImplementedError



"""
The built-in filters.
"""

import os
import sys
import distutils
import subprocess

from teapot.filters.decorators import named_filter


@named_filter('windows')
def windows():
    """
    Check if the platform is windows.
    """

    return sys.platform.startswith('win32')

@named_filter('linux')
def linux():
    """
    Check if the platform is windows.
    """

    return sys.platform.startswith('linux')

@named_filter('darwin')
def darwin():
    """
    Check if the platform is darwin.
    """

    return sys.platform.startswith('darwin')

@named_filter('unix')
def unix():
    """
    Check if the platform is unix.
    """

    return linux() or darwin()

@named_filter('msvc', depends='windows')
def msvc():
    """
    Check if MSVC is available.
    """

    return 'VCINSTALLDIR' in os.environ

@named_filter('msvc-x86', depends='msvc')
def msvc_x86():
    """
    Check if MSVC x86 is available.
    """

    output = subprocess.check_output('cl.exe', shell=True)
    first_line = output.split('\n')[0]

    return first_line.endswith('for x86')

@named_filter('msvc-x64', depends='msvc')
def msvc_x64():
    """
    Check if MSVC x64 is available.
    """

    output = subprocess.check_output('cl.exe', shell=True)
    first_line = output.split('\n')[0]

    return first_line.endswith('for x64')

@named_filter('mingw', depends='windows')
def mingw():
    """
    Check if MinGW is available.
    """

    return bool(distutils.spawn.find_executable('gcc'))



"""
A set of fetcher callbacks class.
"""

from progressbar import *


class BaseFetcherCallback(object):

    """
    Derive from this class to create a fetcher callback class.
    """

    def __init__(self, fetcher):
        """
        Create a new fetcher callback instance.

        `fetcher` is the fetcher instance to be associated with.
        """

        self.fetcher = fetcher

    def on_start(self, target, size):
        """
        Call this method when the fetch is about to begin.

        `target` is the user-friendly name for the target being fetched.
        `size` is the size, in bytes, of the archive that is being fetched.

        You may use a None `size` to indicate that the overall size is unknown.
        """

        pass

    def on_update(self, progress):
        """
        Call this method whenever the fetch had some progress.

        `progress` is the number of bytes that were already fetched.
        """

        pass

    def on_finish(self):
        """
        Call this method when the fetch is over.
        """

        pass

    def on_exception(self, exception):
        """
        Call this method when a fatal exception is raised that must abort the
        fetch.

        The base implementation calls the :func:`on_finish` method and if you
        override this method, you must do the same (or just call the base
        method).
        """

        self.on_finish()


class NullFetcherCallback(BaseFetcherCallback):
    """
    Provides no callback.
    """

    pass


class ProgressBarFetcherCallback(BaseFetcherCallback):
    """
    Provides a progress bar callback to fetchers.
    """

    def on_start(self, target, size):
        """
        The fetch just started.

        You use a None `size` to indicate that the overall size is unknown.
        """

        widgets = [target, ': ', Bar(marker='#', left='[', right='] '), Percentage(), ' - ', FileTransferSpeed()]

        if size:
            widgets.extend([' - ', 'Total size: %s MB' % round(size / (1024 ** 2), 2)])

        self.progressbar = ProgressBar(widgets=widgets, maxval=size)
        self.progressbar.start()

    def on_update(self, progress):
        """
        The fetch was updated.
        """

        try:
            self.progressbar.update(progress)
        except ValueError:
            pass

    def on_finish(self):
        """
        The fetch finished.
        """

        self.progressbar.finish()

    def on_exception(self, exception):
        """
        The fetch failed.
        """

        if hasattr(self, 'progressbar'):
            self.on_finish()



"""
The teapot logging facilities.
"""

import sys
import logging
import colorama

try:
    from logging import NullHandler

except ImportError:
    class NullHandler(logging.Handler):
        """
        A null handler class.
        """

        def emit(self, record):
            pass

LOGGER = logging.getLogger('teapot')
LOGGER.addHandler(NullHandler())

# Add new log levels
def register_log_level(name, value):
    """
    Register a new log level.
    """

    setattr(logging, name.upper(), value)
    logging.addLevelName(value, name.upper())

    def log_func(self, message, *args, **kwargs):
        self._log(value, message, args, **kwargs)

    setattr(logging.Logger, name, log_func)

register_log_level('important', logging.INFO + 1)
register_log_level('success', logging.INFO + 2)

# Extend the LogRecord class getMessage() method to support highlighting
logging.LogRecord.getLegacyMessage = logging.LogRecord.getMessage

def getMessage(self):
    """
    Return the message from this LogRecord.
    """

    colorizer = getattr(self, 'colorizer', None)

    if colorizer:
        self.args = tuple(map(colorizer, self.args))

    return self.getLegacyMessage()

logging.LogRecord.getMessage = getMessage

class Highlight(object):
    """
    Highlight an instance in the logs.
    """

    def __init__(self, msg):
        """
        Create a highlight object.
        """

        self.msg = msg

    def __str__(self):
        """
        Return a string representation.
        """

        return str(self.msg)

    def render(self, color, reset):
        """
        Return a colorized representation.
        """

        return '%s%s%s' % (color, self.msg, reset)

def print_normal(msg):
    """
    Print a normal line.
    """

    sys.stdout.write(colorama.Fore.BLUE + msg + colorama.Style.RESET_ALL)

def print_error(msg):
    """
    Print an error line.
    """

    sys.stdout.write(colorama.Fore.RED + msg + colorama.Style.RESET_ALL)

class ColorizingStreamHandler(logging.StreamHandler):
    """
    A logging handler that colorizes its output.
    """

    COLOR_MAP = {
        logging.DEBUG: colorama.Style.DIM + colorama.Fore.CYAN,
        logging.IMPORTANT: colorama.Style.BRIGHT,
        logging.SUCCESS: colorama.Fore.GREEN,
        logging.WARNING: colorama.Fore.YELLOW,
        logging.ERROR: colorama.Fore.RED,
        logging.CRITICAL: colorama.Back.RED,
    }

    def __init__(self, stream):
        super(ColorizingStreamHandler, self).__init__(
            colorama.AnsiToWin32(stream).stream,
        )

    @property
    def is_tty(self):
        """
        Check if we are in a tty.
        """

        return getattr(self.stream, 'isatty', lambda: False)()

    def format(self, record):
        """
        Format a record.
        """

        if self.is_tty:
            highlight_color = self.COLOR_MAP.get(logging.IMPORTANT)
            log_color = self.COLOR_MAP.get(record.levelno, '')
            reset = colorama.Style.RESET_ALL

            def colorize(msg):

                if isinstance(msg, Highlight):
                    return msg.render(
                        color=highlight_color,
                        reset=reset + log_color,
                    )

                return msg

            record.colorizer = colorize

        message = super(ColorizingStreamHandler, self).format(record)

        if self.is_tty:
            message = '%s%s%s' % (log_color, message, reset)

        return message



"""
teapot 'attendee' class.
"""

import os
import json
import hashlib
import errno
import shutil
import itertools

from contextlib import contextmanager

from teapot.log import LOGGER
from teapot.log import Highlight as hl
from teapot.source import make_sources
from teapot.path import mkdir, rmdir
from teapot.unpackers import get_unpacker_class_for_type
from teapot.builders import make_builders
from teapot.filters import Filtered


def make_attendees(party, data):
    """
    Build a list of attendees from an attendees data dictionary.
    """

    return [
        make_attendee(party, name, attributes)
        for name, attributes
        in data.items()
    ]


def make_attendee(party, name, attributes):
    """
    Create an attendee from a name a dictionary of attributes.
    """

    if not attributes:
        attributes = {}

    prefix = attributes.get('prefix')

    if prefix is True:
        prefix = name

    attendee = Attendee(
        party=party,
        name=name,
        depends=make_depends(attributes.get('depends')),
        filters=attributes.get('filters'),
        prefix=prefix,
    )

    attendee.sources = make_sources(attendee, attributes.get('source'))
    attendee.builders = make_builders(attendee, attributes.get('builders'))

    return attendee


def make_depends(depends):
    """
    Create a list of dependencies.

    `depends` can either be a single dependency name or a list of dependencies.

    If `depends` is False, an empty list is returned.
    """

    if not depends:
        return []

    elif isinstance(depends, basestring):
        return [depends]

    return depends


class NoSuchBuilderError(ValueError):

    """
    The specified builder does not exist.
    """

    def __init__(self, name=None, tag=None):
        """
        Create an NoSuchBuilderError for the specified `name` or `tag`.
        """

        assert name or tag

        if name:
            super(NoSuchBuilderError, self).__init__(
                'No builder found with that name: %s' % name
            )
        else:
            super(NoSuchBuilderError, self).__init__(
                'No builder found with that tag: %s' % tag
            )

        self.name = name
        self.tag = tag


class Attendee(Filtered):

    """
    An `Attendee` instance holds information about an attendee (third-party
    software).
    """

    CACHE_FILE = 'cache.json'
    BUILD_FILE = 'build.json'

    def __init__(self, party, name, depends=[], filters=[], prefix=None):
        """
        Create an attendee associated to a `party`.

        `name` is the name of the attendee.
        `depends` is a list of Attendee names to depend on.
        `filters` is the list of filters that must match for this attendee to
        be active.
        `prefix` is the prefix to use for that attendee.
        """

        if not party:
            raise ValueError('An attendee must be associated to a party.')

        self.party = party
        self.name = name
        self.sources = []
        self.builders = []
        self.depends = set(depends)
        self.prefix = prefix or ''

        Filtered.__init__(self, filters=filters)

    @property
    def enabled_sources(self):
        """
        Get the list of the enabled sources.
        """

        return [source for source in self.sources if source.enabled]

    @property
    def enabled_builders(self):
        """
        Get the list of the enabled builders.
        """

        return [builder for builder in self.builders if builder.enabled]

    def __unicode__(self):
        """
        Get a unicode representation of the attendee.
        """

        return self.name

    def __str__(self):
        """
        Get the name of the attendee.
        """

        return self.name

    def __repr__(self):
        """
        Get a representation of the source.
        """

        return '<%s.%s(party=%r, name=%r, sources=%r, depends=%r)>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.party,
            self.name,
            self.sources,
            self.depends,
        )

    def get_builder_by_name(self, name):
        """
        Get all the builders that match the specified name.

        If no builder has the specified name, a NoSuchBuilderError is raised.
        """

        result = [builder for builder in self.builders if builder.name == name]

        if not result:
            raise NoSuchBuilderError(name=name)

        return result[0]

    def get_builders_by_tag(self, tag):
        """
        Get all the enabled builders that match the specified tag.

        If no builder has the specified tag, a NoSuchBuilderError is raised.
        """

        result = [builder for builder in self.enabled_builders if tag in builder.tags]

        if not result:
            raise NoSuchBuilderError(tag=tag)

        return result

    @property
    def cache_path(self):
        """
        The path of the cache of this attendee.
        """

        return os.path.join(self.party.cache_path, self.name)

    def clean_cache(self):
        """
        Clean the cache directory.
        """

        LOGGER.info('Cleaning cache directory: %s', hl(self.build_path))

        rmdir(self.cache_path)

        LOGGER.info('Done cleaning cache directory for %s.', hl(self))

    def create_cache(self):
        """
        Create the cache directory.
        """

        mkdir(self.cache_path)

    @property
    def build_path(self):
        """
        The path of the source of this attendee.
        """

        return os.path.join(self.party.build_path, self.name)

    @property
    def variant_builds_path(self):
        """
        The path to the variant builds.
        """

        return os.path.join(self.build_path, 'builds')

    @property
    def logs_path(self):
        """
        The path to the log files.
        """

        return os.path.join(self.build_path, 'logs')

    def clean_build(self):
        """
        Clean the build directory.
        """

        LOGGER.info('Cleaning build directory: %s', hl(self.build_path))

        rmdir(self.build_path)

        LOGGER.info('Done cleaning build directory for %s.', hl(self))

    def clean_install(self):
        """
        Clean the install files.
        """

        LOGGER.info('Cleaning install files: %s', hl(self.build_path))

        for builder in self.enabled_builders:
            builder.clean()

        LOGGER.info('Done cleaning install files for %s.', hl(self))

    def fetch(self):
        """
        Fetch the attendee archive by trying all its sources.

        If the fetching suceeds, the succeeding source is returned.
        If the fetching fails, a RuntimeError is raised.
        """

        self.create_cache()

        LOGGER.info('Fetching %s...', hl(self))

        if not self.enabled_sources:
            raise RuntimeError('No active source found for %s' % self.name)

        for source in self.enabled_sources:
            LOGGER.info('Trying from %s...', hl(source))

            cache_info = source.fetch(root_path=self.cache_path)

            if cache_info:
                LOGGER.success('%s fetched successfully.', hl(self))

                self.write_cache_info(cache_info)

                return source

        raise RuntimeError('All sources failed for %s' % self.name)

    def unpack(self):
        """
        Unpack the attendee archive.

        If the unpacking suceeds, the archive source path is returned.
        """

        mkdir(self.build_path)

        LOGGER.info('Unpacking %s...', hl(self))

        build_info = get_unpacker_class_for_type(self.archive_type)(attendee=self).unpack()

        if build_info:
            LOGGER.success('%s unpacked successfully at: %s', hl(self), hl(build_info.get('source_tree_path')))
            build_info['source_tree_origin_hash'] = self.archive_hash

            self.write_build_info(build_info)

    def build(self, tags=None, verbose=False, keep_builds=False, force=False):
        """
        Build the attendee, with the builders that match `tags`, if any.

        If `verbose` is specified, the build output will be printed on stdout
        as the build progresses.

        If `keep_builds` is specified, the build directory won't be erased
        until the next rebuild.

        If `force` is truthy, all builds will be done, no matter what their
        previous status is.
        """

        if not tags:
            builders = self.enabled_builders
        else:
            builders = list(itertools.chain(*map(self.get_builders_by_tag, tags)))

        mkdir(self.build_path)
        mkdir(self.logs_path)

        if not builders:
            LOGGER.warning('Not building %s because no builder matches the current settings. Did you forget to set a filter on the attendee ?', hl(self))
        else:
            LOGGER.info('Building %s with %s builder(s)...', hl(self), len(builders))

            build_info = self.build_info
            build_info.setdefault('builds', {})

            attendee_signature = self.signature
            saved_attendee_signature = build_info.get('signature')

            for builder in builders:
                try:
                    signature = build_info['builds'].get(builder.name)

                    if (attendee_signature == saved_attendee_signature) and (signature == builder.signature) and not force:
                        LOGGER.info('Last build for %s was successful. Not rebuilding unless force rebuild is requested.', hl(builder.name))

                    else:
                        if attendee_signature != saved_attendee_signature:
                            LOGGER.info('Signature for %s changed since last successful build. Building using "%s"...', hl(self), hl(builder))
                        elif not signature:
                            LOGGER.info('Starting first build for %s using builder "%s"...', hl(self), hl(builder))
                        elif not force:
                            LOGGER.info('Build signature changed: starting rebuild for %s using builder "%s"...', hl(self), hl(builder))
                        else:
                            LOGGER.info('Forcing rebuild for %s using builder "%s"...', hl(self), hl(builder))

                            build_info['builds'][builder.name] = None
                            self.write_build_info(build_info)

                        with self.create_temporary_build_directory(builder, persistent=keep_builds) as build_directory:
                            with self.create_log_file(builder) as log_file:
                                builder.build(
                                    build_directory=build_directory,
                                    log_file=log_file,
                                    verbose=verbose,
                                )

                                build_info['builds'][builder.name] = builder.signature
                                self.write_build_info(build_info)

                except Exception as ex:
                    LOGGER.error('Error while building %s: %s', hl(self), ex)

                    raise

                LOGGER.success('Builder "%s" was successful.', hl(builder))

            build_info['signature'] = attendee_signature
            self.write_build_info(build_info)

            LOGGER.success('%s built successfully.', hl(self))

    @contextmanager
    def create_log_file(self, builder):
        """
        Create a log file object and returns it.

        `builder` is the builder that will write to the log file.
        """

        log_file = os.path.join(self.logs_path, '%s.txt' % builder.name)

        try:
            yield open(log_file, 'w')

        finally:
            LOGGER.info('Log file written to: %s', hl(log_file))

    @contextmanager
    def create_temporary_build_directory(self, builder, persistent=False):
        """
        Create a build directory that is a clone of the real build directory.

        The directory name and location depend on the `builder` that will use
        it.

        The directory will be deleted, unless `persistent` is truthy.
        """

        build_directory = os.path.join(self.variant_builds_path, builder.name)

        try:
            rmdir(build_directory)

            LOGGER.info('Copying source tree to %s...', hl(build_directory))
            shutil.copytree(self.source_tree_path, build_directory)

            yield build_directory

        finally:
            if not persistent:
                rmdir(build_directory)
            else:
                LOGGER.info('Keeping build directory at %s', hl(build_directory))

    @property
    def cache_info(self):
        """
        Get the associated cache info.

        Returns a dict containing the archive cache information.

        If the archive cache information does not exist, an empty dict
        is returned.
        """

        try:
            with open(os.path.join(self.cache_path, self.CACHE_FILE)) as cache_file:
                return json.load(cache_file)

        except IOError as ex:
            if ex.errno != errno.ENOENT:
                raise
        except ValueError:
            pass

        return {}

    def write_cache_info(self, value):
        """
        Write the associated cache info.
        """

        with open(os.path.join(self.cache_path, self.CACHE_FILE), 'w') as cache_file:
            return json.dump(value, cache_file)

    @property
    def build_info(self):
        """
        Get the associated build info.

        Returns a dict containing the archive build information.

        If the archive build information does not exist, an empty dict
        is returned.
        """

        try:
            with open(os.path.join(self.build_path, self.BUILD_FILE)) as build_file:
                return json.load(build_file)

        except IOError as ex:
            if ex.errno != errno.ENOENT:
                raise
        except ValueError:
            pass

        return {}

    def write_build_info(self, value):
        """
        Set the associated build info.
        """

        with open(os.path.join(self.build_path, self.BUILD_FILE), 'w') as build_file:
            return json.dump(value, build_file)

    @property
    def source_hash(self):
        """
        Get the source hash.
        """

        return self.cache_info.get('source_hash')

    @property
    def archive_path(self):
        """
        Get the archive path.
        """

        return self.cache_info.get('archive_path')

    @property
    def archive_type(self):
        """
        Get the archive type.
        """

        result = self.cache_info.get('archive_type')

        if result:
            return tuple(result)

    @property
    def archive_hash(self):
        """
        Get the archive hash.
        """

        return self.cache_info.get('archive_hash')

    @property
    def fetched(self):
        """
        Check if the attendee was fetched already.
        """

        if self.archive_path and os.path.exists(self.archive_path):
            return True

    @property
    def source_hash_matches(self):
        """
        Check if one of source hashes match the one in the cache.
        """

        for source in self.enabled_sources:
            if source.source_hash == self.source_hash:
                return True

        return False

    @property
    def source_tree_path(self):
        """
        Get the source tree path.
        """

        return self.build_info.get('source_tree_path')

    @property
    def source_tree_origin_hash(self):
        """
        Get the hash of the archive that was used to create this source tree.
        """

        return self.build_info.get('source_tree_origin_hash')

    @property
    def unpacked(self):
        """
        Check if the attendee needs unpacking.
        """

        if self.source_tree_path and os.path.isdir(self.source_tree_path):
            return True

    @property
    def archive_hash_matches(self):
        """
        Check if the unpacked archive hash matches the one in the cache.
        """

        return self.archive_hash == self.source_tree_origin_hash

    @property
    def signature(self):
        """
        Get the attendee signature.
        """

        LOGGER.debug('Computing source hash for %s...', hl(self))

        data = {
            'sources': [source.source_hash for source in self.sources],
            'depends': [self.party.get_attendee_by_name(depend).signature for depend in self.depends],
            'builders': [builder.signature for builder in self.builders],
        }

        algorithm = hashlib.sha1()
        algorithm.update(json.dumps(data))

        result = algorithm.hexdigest()

        LOGGER.debug('Hash is: %s', hl(result))

        return result



"""
Teapot environments.
"""

import os
import shlex

from teapot.environments.environment_register import EnvironmentRegister
from teapot.environments.environment import Environment


DEFAULT_ENVIRONMENT_NAME = 'default'

def create_default_environment():
    """
    Create a default environment, holding the current process environment
    variables.
    """

    return Environment(name='Default environment', variables=os.environ.copy())

def make_environments(register, environments):
    """
    Make a list of environments from a dictionary.

    If `environments` is a dictionary, make_environment() is called on every
    element, and the result is appended the returned value.

    If `environments` is falsy, an empty list is returned.
    """

    if environments:
        return [
            make_environment(register, name, environment)
            for name, environment
            in environments.items()
        ]

    return []

def make_environment(register, name, environment):
    """
    Make an environment from a dictionary of its attributes.

    `name`, is the name of the environment. If `name` is None, the environment
    is anonymous and won't be registered.

    `environment` is a dictionary of attributes.
    """

    inherit = environment.get('inherit')

    if inherit:
        if isinstance(inherit, basestring):
            inherit = register.get_environment_by_name(inherit)
        else:
            inherit = make_environment(register, None, inherit)

    shell = environment.get('shell', True)

    if shell is not None:
        if isinstance(shell, basestring):
            shell = shlex.split(shell)

    result = Environment(
        name=name,
        variables=environment.get('variables', {}),
        inherit=inherit,
        shell=shell,
    )

    if name:
        register.register_environment(name, result)

    return result



"""
Contains all teapot fetchers logic.

If you want to create your own fetcher, derive from
teapot.fetchers.base_fetcher.BaseFetcher.
"""

from teapot.fetchers.base_fetcher import BaseFetcher, UnsupportedLocationError
from teapot.fetchers.file_fetcher import FileFetcher  # NOQA
from teapot.fetchers.http_fetcher import HttpFetcher  # NOQA
from teapot.fetchers.github_fetcher import GithubFetcher  # NOQA


class NoSuchFetcherError(ValueError):

    """
    Raised when a non-existing fetcher was requested.
    """

    def __init__(self, shortname):
        """
        Create a NoSuchFetcherError instance for the specified `shortname`.
        """

        super(
            NoSuchFetcherError,
            self
        ).__init__(
            'No such fetcher (%r)' % shortname
        )


def get_fetcher_class_from_shortname(shortname):
    """
    Get the appropriate fetcher class/builder from its `shortname`.

    If `shortname` is false, the automatic fetcher builder is returned. See
    guess_fetcher_instance().

    If `shortname` does not match with an existing fetcher shortname, a
    NoSuchFetcherError instance is raised.
    """

    if not shortname or shortname == 'auto':
        return guess_fetcher_instance

    fetcher_class = BaseFetcher.index.get(shortname)

    if fetcher_class:
        return fetcher_class

    raise NoSuchFetcherError(shortname=shortname)


def guess_fetcher_instance(source, default_class=None):
    """
    Guess the fetcher class to use for the specified `source` and returns an
    instance of this class.

    `default_class` can be either a fetcher class or a fetcher shortname to use
    a default if no guess can be made from the specified location.

    If `default_class` is false and no guess can be made, no instance is
    returned.
    """

    if default_class and isinstance(default_class, basestring):
        default_class = get_fetcher_class_from_shortname(default_class)

        if default_class == guess_fetcher_instance:
            raise ValueError(
                'Specifying the auto fetcher class as the default type is not allowed as it could lead to infinite recursion.'
            )

    # We iterate through the registered fetchers classes and try to instantiate
    # one with the location to detect if it is supported.
    for fetcher_class in BaseFetcher.index.values():
        try:
            return fetcher_class(source)

        except UnsupportedLocationError:
            pass

    if default_class:
        return default_class(source)



"""
The teapot main script.
"""

import os
import sys
import argparse
import logging

from functools import wraps

from teapot.log import LOGGER, ColorizingStreamHandler
from teapot.party import load_party_file


def main():
    """
    The script entry point.
    """

    parser = argparse.ArgumentParser(
        description='Manage third-party software.'
    )

    parser.add_argument('-d', '--debug', action='store_true', help='Enable debug output.')
    parser.add_argument('-v', '--verbose', action='store_true', help='Be more explicit about what happens.')

    command_parser = parser.add_subparsers(help='The available commands.')
    parser.add_argument('-p', '--party-file', default=None, help='The party-file to read.')

    # The clean command
    clean_command_parser = command_parser.add_parser('clean', help='Clean the party.')
    clean_subcommand_parser = clean_command_parser.add_subparsers(help='The available commands.')

    #  The clean cache subcommand
    clean_cache_command_parser = clean_subcommand_parser.add_parser('cache', help='Clean the party cache.')
    clean_cache_command_parser.set_defaults(func=clean_cache)
    clean_cache_command_parser.add_argument('attendees', metavar='attendee', nargs='*', default=[], help='The attendees to clean.')

    #  The clean build subcommand
    clean_build_command_parser = clean_subcommand_parser.add_parser('build', help='Clean the party build.')
    clean_build_command_parser.set_defaults(func=clean_build)
    clean_build_command_parser.add_argument('attendees', metavar='attendee', nargs='*', default=[], help='The attendees to clean.')

    #  The clean install subcommand
    clean_install_command_parser = clean_subcommand_parser.add_parser('install', help='Clean the party install.')
    clean_install_command_parser.set_defaults(func=clean_install)
    clean_install_command_parser.add_argument('attendees', metavar='attendee', nargs='*', default=[], help='The attendees to clean.')

    #  The clean all subcommand
    clean_all_command_parser = clean_subcommand_parser.add_parser('all', help='Clean the party cache, build and install.')
    clean_all_command_parser.set_defaults(func=clean_all)
    clean_all_command_parser.add_argument('attendees', metavar='attendee', nargs='*', default=[], help='The attendees to clean.')

    # The fetch command
    fetch_command_parser = command_parser.add_parser('fetch', help='Fetch all the archives.')
    fetch_command_parser.set_defaults(func=fetch)
    fetch_command_parser.add_argument('attendees', metavar='attendee', nargs='*', default=[], help='The attendees to fetch.')
    fetch_command_parser.add_argument('-f', '--force', action='store_true', help='Fetch archives even if they already exist in the cache.')

    # The unpack command
    unpack_command_parser = command_parser.add_parser('unpack', help='Unpack all the fetched archives.')
    unpack_command_parser.set_defaults(func=unpack)
    unpack_command_parser.add_argument('attendees', metavar='attendee', nargs='*', default=[], help='The attendees to unpack.')
    unpack_command_parser.add_argument('-f', '--force', action='store_true', help='Unpack archives even if they already exist in the build.')

    # The build command
    build_command_parser = command_parser.add_parser('build', help='Build the archives.')
    build_command_parser.set_defaults(func=build)
    build_command_parser.add_argument('attendees', metavar='attendee', nargs='*', default=[], help='The attendees to build.')
    build_command_parser.add_argument('-t', '--tags', metavar='tag', action='append', default=[], help='The tags to build.')
    build_command_parser.add_argument('-u', '--force-unpack', action='store_true', help='Delete and reunpack all source tree directories before attempting a build.')
    build_command_parser.add_argument('-f', '--force-build', action='store_true', help='Run all builders even if their last run was successful.')
    build_command_parser.add_argument('-k', '--keep-builds', action='store_true', help='Keep the build directories for inspection.')

    args = parser.parse_args()

    handler = ColorizingStreamHandler(sys.stdout)
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)

    if args.debug:
        logging.getLogger().addHandler(handler)
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        LOGGER.addHandler(handler)
        LOGGER.setLevel(logging.INFO)

    if args.party_file is None:
        args.party_file = os.path.join(os.getcwd(), 'party.yaml')

    try:
        party = load_party_file(args.party_file)

    except IOError:
        LOGGER.error('No party-file was found. (Searched path: "%s")', args.party_file)

        return 1

    if not args.func(party, args):
        return 2

def command(func):
    """
    Provides exception handling for commands.
    """

    @wraps(func)
    def decorated(party, args, **kwargs):
        try:
            func(party, args, **kwargs)

            return True

        except KeyboardInterrupt:
            pass

        except Exception as ex:
            if args.debug:
                LOGGER.exception(ex)
            else:
                LOGGER.error(str(ex))

        LOGGER.error('teapot execution failed.')

        return False

    return decorated

@command
def clean_cache(party, args):
    """
    Clean the party cache.
    """

    party.clean_cache(
        attendees=args.attendees,
    )

@command
def clean_build(party, args):
    """
    Clean the party build.
    """

    party.clean_build(
        attendees=args.attendees,
    )

@command
def clean_install(party, args):
    """
    Clean the party install.
    """

    party.clean_install(
        attendees=args.attendees,
    )

@command
def clean_all(party, args):
    """
    Clean the party cache and build.
    """

    party.clean_cache(
        attendees=args.attendees,
    )
    party.clean_build(
        attendees=args.attendees,
    )
    party.clean_install(
        attendees=args.attendees,
    )

@command
def fetch(party, args):
    """
    Fetch the archives.
    """

    party.fetch(
        attendees=args.attendees,
        force=args.force,
    )

@command
def unpack(party, args):
    """
    Unpack the archives.
    """

    party.unpack(
        attendees=args.attendees,
        force=args.force,
    )

@command
def build(party, args):
    """
    Build the archives.
    """

    party.build(
        attendees=args.attendees,
        tags=args.tags,
        verbose=args.verbose,
        force_unpack=args.force_unpack,
        keep_builds=args.keep_builds,
        force_build=args.force_build,
    )



"""
Contains all teapot unpackers logic.

If you want to create your own unpacker, derive from
teapot.unpackers.base_unpacker.BaseUnpacker.
"""

from teapot.unpackers.base_unpacker import BaseUnpacker
from teapot.unpackers.null_unpacker import NullUnpacker
from teapot.unpackers.tarball_unpacker import TarballUnpacker
from teapot.unpackers.zipfile_unpacker import ZipFileUnpacker


class NoSuchUnpackerError(ValueError):

    """
    Raised when a non-existing unpacker was requested.
    """

    def __init__(self, _type):
        """
        Create a NoSuchUnpackerError instance for the specified `_type`.
        """

        super(
            NoSuchUnpackerError,
            self
        ).__init__(
            'No suitable unpacker found for: %s' % repr(_type)
        )


def get_unpacker_class_for_type(_type):
    """
    Get the appropriate unpacker class/builder for a given `_type`.

    If no unpacker supports `_type` a NoSuchUnpackerError instance is
    raised.
    """

    if not _type:
        raise ValueError('Unable to get an unpacker as no type was specified.')

    unpacker_class = BaseUnpacker.index.get(tuple(_type))

    if unpacker_class:
        return unpacker_class

    raise NoSuchUnpackerError(_type=_type)



"""
A teapot path-handling class.
"""

import os
import stat
import shutil
import errno

from teapot.log import LOGGER


def from_user_path(path):
    """
    Perform all variables substitutions from the specified user path.
    """

    return os.path.normpath(os.path.expanduser(os.path.expandvars(path)))

def read_path(value, base_path, default_path):
    """
    Read a path value from a string.

    If `value` is a string, the Cache is default created at the location
    specified by `value`.

    If `value` is falsy, the Cache is default created.
    """

    if not value:
        cache_path = from_user_path(default_path)
    else:
        cache_path = from_user_path(value)

    if not os.path.isabs(cache_path):
        cache_path = os.path.normpath(os.path.join(base_path, cache_path))

    return cache_path

def mkdir(path):
    """
    Create the specified path.

    Does nothing if the path exists.
    """

    try:
        LOGGER.debug('Creating directory at %s.', path)

        os.makedirs(path)

    except OSError as ex:
        if ex.errno != errno.EEXIST or not os.path.isdir(path):
            raise

def rmdir(path):
    """
    Delete the specified path if it exists.

    Does nothing if the path doesn't exist.
    """

    try:
        LOGGER.debug('Removing directory at %s.', path)

        def onerror(func, path, excinfo):
            if os.path.exists(path):
                LOGGER.debug('Was unable to delete "%s": %s', path, excinfo[1])
                LOGGER.debug('Trying again after changing permissions...')
                os.chmod(path, stat.S_IWUSR)

                try:
                    func(path)
                except Exception as ex:
                    LOGGER.error('Unable to delete "%s": %s', path, excinfo[1])

                    raise

        shutil.rmtree(path, ignore_errors=False, onerror=onerror)

    except Exception as ex:
        LOGGER.debug(ex)

def windows_to_unix_path(path):
    """
    Convert a Windows path to a UNIX path, in such a way that it can be used in
    MSys or Cygwin.
    """

    drive, tail = os.path.splitdrive(path)

    if drive:
        drive = '/' + drive[0]

    return drive + tail.replace('\\', '/')



"""
A local-file fetcher class.
"""

import os
import shutil
import mimetypes

from teapot.fetchers.base_fetcher import BaseFetcher
from teapot.path import rmdir


class FileFetcher(BaseFetcher):

    """
    Fetchs a file on the local filesystem.
    """

    shortname = 'file'

    def read_source(self, source):
        """
        Checks that the `source` is a local filename.
        """

        if os.path.exists(source.location):
            self.file_path = os.path.abspath(source.location)

            return True

    def do_fetch(self, target):
        """
        Fetch a filename.
        """

        archive_path = os.path.join(target, os.path.basename(self.file_path))

        archive_type = mimetypes.guess_type(self.file_path)
        size = os.path.getsize(self.file_path)

        rmdir(archive_path)

        self.progress.on_start(target=os.path.basename(archive_path), size=size)

        if os.path.isfile(self.file_path):
            shutil.copyfile(self.file_path, archive_path)
        elif os.path.isdir(self.file_path):
            shutil.copytree(self.file_path, archive_path)
        else:
            raise RuntimeError('Unsupported path: %s' % self.file_path)

        # No real interactive progress to show here.
        #
        # This could be fixed though.

        self.progress.on_update(progress=size)
        self.progress.on_finish()

        return {
            'archive_path': archive_path,
            'archive_type': archive_type,
        }



"""
The filters decorators.
"""

from functools import wraps


FILTERS = {}

class InvalidFilterError(ValueError):

    """
    No filter exists with the specified name.
    """

    def __init__(self, name):
        """
        Create an InvalidFilterError for the specified filter `name`.
        """

        self.name = name

        super(InvalidFilterError, self).__init__(
            'No filter with the specified name: %s' % name
            )


class DuplicateFilterError(ValueError):

    """
    Another filter with the same name exists.
    """

    def __init__(self, name):
        """
        Create a DuplicateFilterError for the specified filter `name`.
        """

        self.name = name

        super(DuplicateFilterError, self).__init__(
            'Another filter was already registered with the name %r' % name
            )


class named_filter(object):

    """
    Registers a function to be a filter.
    """

    def __init__(self, name, depends=None, override=False):
        """
        Registers the function with the specified name.

        If another function was registered with the same name, a
        `DuplicateFilterError` will be raised, unless `override` is truthy.

        `depends` can be either:
            - `None`, if the filter does not depend on any other filter.
            - A `str` instance, being the registered name of another filter to depend on.
            - A `list` of `str` instances, being the list of registered names of other filters to depend on.

        If a filter depends on other filters, those will be checked *before* the actual filter gets run.
        """

        if name in FILTERS and not override:
            raise DuplicateFilterError(name)

        self.name = name

        if not depends:
            self.depends = []
        elif isinstance(depends, basestring):
            self.depends = [get_filter_by_name(depends)]
        else:
            self.depends = map(get_filter_by_name, depends)

    def __call__(self, func):
        """
        Registers the function and returns it unchanged.
        """

        @wraps(func)
        def decorated(*args, **kwargs):
            return all(f() for f in self.depends + [func])

        FILTERS[self.name] = decorated

        return decorated

def get_filter_by_name(name):
    """
    Get a filter by name.

    If no filter matches the specified name, an InvalidFilterError is raised.
    """

    if not name in FILTERS:
        raise InvalidFilterError(name=name)

    return FILTERS[name]



"""
Contains all teapot builders logic.
"""

import os
import re
import json
import math
import signal
import hashlib
import subprocess

from contextlib import contextmanager
from threading import Thread
from datetime import datetime

from teapot.log import LOGGER, print_normal, print_error
from teapot.log import Highlight as hl
from teapot.filters import Filtered
from teapot.environments import make_environment, DEFAULT_ENVIRONMENT_NAME
from teapot.extensions import parse_extension


def make_builders(attendee, builders):
    """
    Make a list of builders from a dictionary.
    """

    if not builders:
        return []

    return [make_builder(attendee, name, attributes) for name, attributes in builders.items()]

def make_builder(attendee, name, attributes):
    """
    Make a builder from its attributes.
    """

    if not isinstance(attributes, dict):
        raise ValueError('builder attributes must be a dictionary.')

    tags = attributes.get('tags')

    if not tags:
        tags = []
    elif isinstance(tags, basestring):
        tags = [tags]

    commands = attributes.get('commands', [])

    if isinstance(commands, basestring):
        commands = [commands]

    clean_commands = attributes.get('clean_commands', [])

    if isinstance(clean_commands, basestring):
        clean_commands = [clean_commands]

    filters = attributes.get('filters')
    prefix = attributes.get('prefix')

    if prefix is True:
        prefix = name

    environment = attributes.get('environment', DEFAULT_ENVIRONMENT_NAME)

    if environment:
        if isinstance(environment, basestring):
            environment = attendee.party.environment_register.get_environment_by_name(environment)
        else:
            environment = make_environment(attendee.party.environment_register, None, environment)
    else:
        environment = make_environment(attendee.party.environment_register, None, {})

    return Builder(
        attendee=attendee,
        name=name,
        tags=tags,
        commands=commands,
        environment=environment,
        filters=filters,
        directory=attributes.get('directory'),
        prefix=prefix,
        clean_commands=clean_commands,
    )


class Builder(Filtered):

    """
    A Builder represents a way to build an attendee.
    """

    def __init__(self, attendee, name, tags, commands, environment, filters=[], directory=None, prefix=None, clean_commands=[]):
        """
        Initialize a builder attached to `attendee` with the specified `name`.

        A builder may have `tags`, which must be a list of strings.

        You must specify `commands`, a list of commands to call for the build to
        take place.

        `environment` is the environment to use for the build.

        `filters` is a list of filters that must all validate in order for the
        build to be active in the current environment.

        `directory`, if specified, is a directory relative to the source root,
        where to go before issuing the build commands.

        `prefix` is a prefix that will be added to the right of the global
        prefix inside the $PREFIX variable.

        `clean_commands` is a list of commands to run to clean the install.
        Those commands are executed in sequence until the last one, even if one
        of them fails.
        """

        if not commands:
            raise ValueError('A builder must have at least one command.')

        assert environment

        self.attendee = attendee
        self.name = name
        self.tags = tags
        self.commands = commands or []
        self.environment = environment
        self.directory = directory or None
        self.prefix = prefix or ''
        self.clean_commands = clean_commands

        Filtered.__init__(self, filters=filters)

    def __str__(self):
        """
        Get the name of the builder.
        """

        return self.name

    def __repr__(self):
        """
        Get a representation of the builder.
        """

        return '<%s.%s(name=%r, tags=%r, commands=%r)>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.name,
            self.tags,
            self.commands,
        )

    @contextmanager
    def chdir(self, path):
        """
        Change the current directory.
        """

        old_path = os.getcwd()

        LOGGER.debug('Moving to: %s', hl(path))
        os.chdir(path)

        try:
            yield path
        finally:
            LOGGER.debug('Moving back to: %s', hl(old_path))
            os.chdir(old_path)

    def build(self, build_directory, log_file=None, verbose=False):
        """
        Build the attendee.

        `build_directory` is the directory to build into.

        `log_file`, if specified, is a file object to write the output to.

        `verbose` is a flag that, if truthy, cause the build output to be
        printed on standard output as the build progresses.
        """

        if self.directory:
            source_tree_path = os.path.join(build_directory, self.directory)
        else:
            source_tree_path = build_directory

        if log_file:
            log_file.write('Build started in %s at %s.\n' % (source_tree_path, datetime.now().strftime('%c')))

        with self.chdir(source_tree_path):

            with self.environment.enable() as env:
                if env.shell:
                    LOGGER.info('Building within: %s', hl(' '.join(env.shell)))
                else:
                    LOGGER.info('Building within %s.', hl('the default system shell'))

                if log_file:
                    if env.shell:
                        log_file.write('Using "%s" as a shell.\n' % ' '.join(env.shell))
                    else:
                        log_file.write('Using system shell.\n')

                for key, value in os.environ.iteritems():
                    LOGGER.debug('%s: %s', key, hl(value))

                for index, command in enumerate(self.commands):
                    command = self.apply_extensions(command)
                    numbered_prefix = ('%%0%sd' % int(math.ceil(math.log10(len(self.commands))))) % index

                    LOGGER.important('%s: %s', numbered_prefix, hl(command))

                    if log_file:
                        log_file.write('%s: %s\n' % (numbered_prefix, command))

                    if env.shell:
                        process = subprocess.Popen(env.shell + [command], shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    else:
                        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

                    def handler(signum, frame):
                        LOGGER.warning('The building process was interrupted by the user.')
                        process.terminate()

                    previous_handler = signal.signal(signal.SIGINT, handler)

                    mixed_output = []

                    try:
                        def read_stdout():
                            for line in iter(process.stdout.readline, ''):
                                mixed_output.append((print_normal, line))

                                if log_file:
                                    log_file.write(line)

                                if verbose:
                                    print_normal(line)

                        def read_stderr():
                            for line in iter(process.stderr.readline, ''):
                                mixed_output.append((print_error, line))

                                if log_file:
                                    log_file.write(line)

                                if verbose:
                                    print_error(line)

                        stdout_thread = Thread(target=read_stdout)
                        stdout_thread.daemon = True
                        stdout_thread.start()

                        stderr_thread = Thread(target=read_stderr)
                        stderr_thread.daemon = True
                        stderr_thread.start()

                        stdout_thread.join()
                        stderr_thread.join()

                        process.wait()

                    finally:
                        signal.signal(signal.SIGINT, previous_handler)

                    if log_file:
                        log_file.write('\n')

                    if process.returncode != 0:
                        if not verbose:
                            for func, line in mixed_output:
                                func(line)

                        if log_file:
                            log_file.write('Command failed with status: %s\n' % process.returncode)
                            log_file.write('Build failed at %s.\n' % datetime.now().strftime('%c'))

                        raise subprocess.CalledProcessError(returncode=process.returncode, cmd=command)

        if log_file:
            log_file.write('Build succeeded at %s.\n' % datetime.now().strftime('%c'))

    def clean(self):
        """
        Clean the attendee.
        """

        if not self.clean_commands:
            LOGGER.info('Cleaning builder %s: no clean commands specified. Doing nothing.', hl(self))

        else:
            LOGGER.info('Cleaning builder: %s', hl(self))

            with self.environment.enable() as env:
                if env.shell:
                    LOGGER.info('Executing clean commands within: %s', hl(' '.join(env.shell)))
                else:
                    LOGGER.info('Executing clean commands within %s.', hl('the default system shell'))

                for key, value in os.environ.iteritems():
                    LOGGER.debug('%s: %s', key, hl(value))

                for index, command in enumerate(self.clean_commands):
                    command = self.apply_extensions(command)
                    numbered_prefix = ('%%0%sd' % int(math.ceil(math.log10(len(self.commands))))) % index

                    LOGGER.important('%s: %s', numbered_prefix, hl(command))

                    if env.shell:
                        process = subprocess.Popen(env.shell + [command], shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    else:
                        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

                    def handler(signum, frame):
                        LOGGER.warning('The cleaning process was interrupted by the user.')
                        process.terminate()

                    previous_handler = signal.signal(signal.SIGINT, handler)

                    try:
                        def read_stdout():
                            for line in iter(process.stdout.readline, ''):
                                LOGGER.debug(line.rstrip())

                        def read_stderr():
                            for line in iter(process.stderr.readline, ''):
                                LOGGER.debug(line.rstrip())

                        stdout_thread = Thread(target=read_stdout)
                        stdout_thread.daemon = True
                        stdout_thread.start()

                        stderr_thread = Thread(target=read_stderr)
                        stderr_thread.daemon = True
                        stderr_thread.start()

                        stdout_thread.join()
                        stderr_thread.join()

                        process.wait()

                    finally:
                        signal.signal(signal.SIGINT, previous_handler)

                    if process.returncode != 0:
                        LOGGER.warning('Command failed with status: %s', process.returncode)

            LOGGER.info('Clean succeeded at %s.', datetime.now().strftime('%c'))

    def apply_extensions(self, command):
        """
        Apply the extensions to the command.
        """

        def replace(match):
            code = match.group('code')

            return str(parse_extension(code, self))

        return re.sub(r'\{{(?P<code>[\w\s()]+)}}', replace, command)

    @property
    def signature(self):
        """
        Get the signature of the builder.
        """

        data = {
            'name': self.name,
            'tags': self.tags,
            'commands': self.commands,
            'environment': self.environment.signature,
            'directory': self.directory,
            'prefix': self.prefix,
        }

        algorithm = hashlib.sha1()
        algorithm.update(json.dumps(data))

        result = algorithm.hexdigest()

        LOGGER.debug('%s\'s signature is: %s', hl(self), hl(result))

        return result



"""
teapot 'party' class.
"""

import os
import imp
import yaml

from functools import wraps

from teapot.log import LOGGER
from teapot.log import Highlight as hl
from teapot.attendee import Attendee, make_attendees
from teapot.path import read_path, rmdir
from teapot.defaults import *
from teapot.fetchers.callbacks import ProgressBarFetcherCallback
from teapot.unpackers.callbacks import ProgressBarUnpackerCallback
from teapot.environments import make_environments, Environment, EnvironmentRegister, create_default_environment, DEFAULT_ENVIRONMENT_NAME


def load_party_file(path):
    """
    Create a Party instance from a party-file.

    `path` must be a valid party-file name.
    """

    LOGGER.debug('Opening party-file at %s...', hl(path))

    with open(path) as party_file:
        data = party_file.read()

    values = yaml.load(data)

    extension_modules = values.get('extension_modules', {})

    # We import the extension modules.
    #
    # This has to be done first, or the things defined in those extension
    # modules won't be accessible to the parsing functions.

    modules = []

    for name, module_path in extension_modules.items():
        abs_module_path = os.path.join(os.path.dirname(path), module_path)

        try:
            modules.append(imp.load_source(name, abs_module_path))
        except Exception as ex:
            LOGGER.error('Unable to load the extension module "%s" at "%s": %s', hl(name), hl(abs_module_path), ex)

            raise

    party = Party(
        path=path,
        cache_path=values.get('cache_path'),
        build_path=values.get('build_path'),
        prefix=values.get('prefix'),
    )

    party.environments = make_environments(party.environment_register, values.get('environments'))
    party.attendees = make_attendees(party, values.get('attendees'))
    party.modules = modules

    return party


class NoSuchAttendeeError(ValueError):

    """
    The specified attendee does not exist.
    """

    def __init__(self, name):
        """
        Create an NoSuchAttendeeError for the specified `name`.
        """

        super(NoSuchAttendeeError, self).__init__(
            'No such attendee: %s' % name,
        )

        self.name = name


class CyclicDependencyError(ValueError):

    """
    The dependency graph is cyclic.
    """

    def __init__(self, cycle):
        """
        Create a CyclicDependencyError for the specified cycle.
        """

        super(CyclicDependencyError, self).__init__(
            'The attendee dependency graph is cyclic: %s' % ' -> '.join(map(str, cycle)),
        )

        self.cycle = cycle


def has_attendees(func):
    """
    Ensures the function has real attendees instances as its `attendees`
    parameter.
    """

    @wraps(func)
    def result(self, *args, **kwargs):
        attendees = kwargs.get('attendees', [])

        if not attendees:
            attendees = self.enabled_attendees
        else:
            attendees = map(self.get_attendee_by_name, attendees)

        kwargs['attendees'] = attendees

        return func(self, *args, **kwargs)

    return result

def has_ordered_attendees(func):
    """
    Ensures the function gets a list of real attendees instances, ordered, with
    less dependant attendees first.
    """

    @wraps(func)
    def result(self, *args, **kwargs):
        attendees = kwargs.get('attendees', [])

        kwargs['attendees'] = self.get_ordered_attendees(attendees=attendees)

        return func(self, *args, **kwargs)

    return result

class Party(object):

    """
    A party object is the root object that stores all the information about the
    different attendees (third-party softwares), and the party options.
    """

    POST_ACTIONS = []

    @classmethod
    def register_post_action(cls, action):
        """
        Register a post action.

        Post action are called in the registration order, every time a new
        `Party` instance is created.
        """

        cls.POST_ACTIONS.append(action)

        return action

    def __init__(self, path=None, cache_path=None, build_path=None, prefix=None, **kwargs):
        """
        Create a Party instance.

        `path` is the path to the party file.
        `cache_path` is the root of the cache.
        `build_path` is the root of the build.
        `prefix` is the prefix common to all attendees.
        """

        self.path = os.path.abspath(path or os.getcwd())
        self.attendees = []
        self.environment_register = EnvironmentRegister()
        self.cache_path = read_path(cache_path, os.path.dirname(self.path), DEFAULT_CACHE_PATH)
        self.build_path = read_path(build_path, os.path.dirname(self.path), DEFAULT_BUILD_PATH)
        self.prefix = read_path(prefix, os.path.dirname(self.path), DEFAULT_PREFIX)
        self.auto_fetch = True
        self.fetcher_callback_class = ProgressBarFetcherCallback
        self.unpacker_callback_class = ProgressBarUnpackerCallback
        self.modules = []

        for post_action in self.POST_ACTIONS:
            post_action(self)

    @property
    def enabled_attendees(self):
        """
        Get the list of the enabled attendees.
        """

        return [attendee for attendee in self.attendees if attendee.enabled]

    def get_attendee_by_name(self, name):
        """
        Get an attendee by name, if it exists.

        If no attendee has the specified name, a NoSuchAttendeeError is raised.
        """

        if isinstance(name, Attendee):
            return name

        for attendee in self.attendees:
            if attendee.name == name:
                return attendee

        raise NoSuchAttendeeError(name=str(name))

    @has_attendees
    def get_ordered_attendees(self, attendees=[]):
        """
        Get a list of attendees ordered in such a way that the less dependent
        attendees start the list.
        """

        graph = {}

        def populate_graph(graph, attendees):

            for attendee in attendees:
                if not attendee.name in graph:
                    graph[attendee.name] = set(attendee.depends)
                    populate_graph(graph, map(self.get_attendee_by_name, attendee.depends))

        populate_graph(graph, attendees)

        ordered_attendees = []

        while graph:

            try:
                attendee = next(node[0] for node in graph.iteritems() if not node[1])

            except StopIteration:
                # We have a cyclic graph. Raise an error and fail.

                cycle = []
                cyclic_graph = graph.copy()

                while cyclic_graph:
                    attendee = next(cyclic_graph.iteritems())[0]

                    if attendee in cycle:
                        break
                    else:
                        cycle.append(attendee)
                        del cyclic_graph[attendee]

                # Reduce the cycle to the smallest sequence.
                cycle = cycle[cycle.index(list(graph[attendee])[0]):]

                cycle = map(self.get_attendee_by_name, cycle)

                raise CyclicDependencyError(cycle=cycle)

            ordered_attendees.append(self.get_attendee_by_name(attendee))
            del graph[attendee]

            for node, depends in graph.iteritems():
                if attendee in depends:
                    depends.remove(attendee)

        return ordered_attendees

    @has_attendees
    def clean_cache(self, attendees=[]):
        """
        Clean the cache, optionally for a given list of `attendees` to recover
        disk space.
        """

        LOGGER.info('Cleaning the cache directory for: %s', hl(', '.join(map(str, attendees))))

        map(Attendee.clean_cache, attendees)

        LOGGER.info('Done cleaning the cache directory.')

    @has_attendees
    def clean_build(self, attendees=[]):
        """
        Clean the build, optionally for a given list of `attendees` to recover
        disk space.
        """

        LOGGER.info('Cleaning the build directory for: %s', hl(', '.join(map(str, attendees))))

        map(Attendee.clean_build, attendees)

        LOGGER.info('Done cleaning the build directory.')

    @has_attendees
    def clean_install(self, attendees=[]):
        """
        Clean the install, optionally for a given list of `attendees` to recover
        disk space.
        """

        LOGGER.info('Cleaning the install files for: %s', hl(', '.join(map(str, attendees))))

        map(Attendee.clean_install, attendees)

        LOGGER.info('Done cleaning the install files.')

    @has_attendees
    def fetch(self, attendees=[], force=False):
        """
        Fetch the archives.
        """

        if force:
            map(lambda x: x.clean_cache(), attendees)

        attendees_to_fetch = [x for x in attendees if not x.fetched or not x.source_hash_matches]

        for attendee in attendees_to_fetch:
            if not attendee.fetched:
                LOGGER.debug('%s was never fetched. Doing it now.', hl(attendee))
            elif not attendee.source_hash_matches:
                LOGGER.info('The source has changed for %s since it was last fetched. Fetching again.', hl(attendee))

        if not attendees_to_fetch:
            if len(attendees) == 1:
                LOGGER.info('%s was already fetched.', hl(attendees[0]))
            else:
                LOGGER.info('None of the %s archives needs fetching.', len(attendees))

        else:
            LOGGER.info("Fetching %s/%s archive(s)...", len(attendees_to_fetch), len(self.enabled_attendees))

            try:
                map(Attendee.fetch, attendees_to_fetch)
            except Exception as ex:
                LOGGER.error('Error: %s', ex)

                raise RuntimeError('Fetching of %s failed' % self)

            LOGGER.info("Done fetching archives.")

    @has_attendees
    def unpack(self, attendees=[], force=False):
        """
        Unpack the archives.
        """

        if self.auto_fetch:
            self.fetch(attendees=attendees)

        if force:
            map(lambda x: x.clean_build(), attendees)

        attendees_to_unpack = [x for x in attendees if not x.unpacked or not x.archive_hash_matches]

        for attendee in attendees_to_unpack:
            if not attendee.unpacked:
                LOGGER.debug('%s was never unpacked. Doing it now.', hl(attendee))
            elif not attendee.archive_hash_matches:
                LOGGER.info('Source tree hash does not match the archive hash for %s. Unpacking again.', hl(attendee))

        if not attendees_to_unpack:
            if len(attendees) == 1:
                LOGGER.info('%s was already unpacked.', hl(attendees[0]))
            else:
                LOGGER.info('None of the %s archive(s) needs unpacking.', len(attendees))

        else:
            LOGGER.info("Unpacking %s/%s archive(s)...", len(attendees_to_unpack), len(self.enabled_attendees))

            map(Attendee.unpack, attendees_to_unpack)

            LOGGER.info("Done unpacking archives.")

    @has_ordered_attendees
    def build(self, attendees=[], tags=[], verbose=False, force_unpack=False, keep_builds=False, force_build=False):
        """
        Build the archives.
        """

        if self.auto_fetch:
            self.unpack(attendees=attendees, force=force_unpack)

        LOGGER.info("Building %s archive(s)...", len(attendees))

        for attendee in attendees:
            attendee.build(tags=tags, verbose=verbose, keep_builds=keep_builds, force=force_build)

        LOGGER.info("Done building archives.")


@Party.register_post_action
def add_default_environment(party):
    """
    Add the default environment to the party.
    """

    party.environment_register.register_environment(DEFAULT_ENVIRONMENT_NAME, create_default_environment())



"""
A Github fetcher class.
"""

import re

from teapot.fetchers.http_fetcher import HttpFetcher


class GithubFetcher(HttpFetcher):

    """
    Fetchs an archive from Github.
    """

    shortname = 'github'

    def read_source(self, source):
        """
        Checks that the `source` is a Github URL.
        """

        assert(source)

        match = re.match(r'github:(?P<owner>[\S/]+)/(?P<repo>[\S/]+)/(?P<ref>.+)', source.location)

        if match:
            values = match.groupdict()
            values['archive_format'] = 'tarball'

            self.url = 'https://api.github.com/repos/%(owner)s/%(repo)s/%(archive_format)s/%(ref)s' % values

            return True



"""
A set of unpacker callbacks class.
"""

import os

from progressbar import *


class BaseUnpackerCallback(object):

    """
    Derive from this class to create an unpacker callback class.
    """

    def __init__(self, unpacker):
        """
        Create a new unpacker callback instance.

        `unpacker` is the unpacker instance to be associated with.
        """

        self.unpacker = unpacker

    def on_start(self, count):
        """
        Call this method when the unpacking is about to start.

        `count` indicates the total count of files in the archive.

        You may use a None `count` to indicate that the overall count is
        unknown.
        """

        pass

    def on_update(self, current_file, progress):
        """
        Call this method when the unpacking had some progress.

        `current_file` indicates the file currently being unpacked.
        `progress` indicates the number of the current file in the archive.
        """

        pass

    def on_finish(self):
        """
        Call this method when the unpacking is complete.
        """

        pass

    def on_exception(self, exception):
        """
        Call this method when a fatal exception is raised that must abort the
        unpack.

        The base implementation calls the :func:`on_finish` method and if you
        override this method, you must do the same (or just call the base
        method).
        """

        self.on_finish()


class NullUnpackerCallback(BaseUnpackerCallback):
    """
    Provides no callback.
    """

    pass


class CurrentFile(WidgetHFill):
    """Displays the current file."""

    def __init__(self, unpacker_callback, prefix='', suffix=''):
        """
        Create a current file widget.
        """

        self.unpacker_callback = unpacker_callback
        self.count = 0
        self.prefix = prefix
        self.suffix = suffix

    def update(self, pbar, width):
        """
        Update the progress bar.
        """

        if self.unpacker_callback.current_file:
            result = self.prefix + self.unpacker_callback.current_file + self.suffix
        else:
            result = ''

        if len(result) > width:
            result = result[:width - 3 - len(self.suffix)] + '...' + self.suffix
        else:
            result += ' ' * (width - len(result))

        return result


class ProgressBarUnpackerCallback(BaseUnpackerCallback):
    """
    Provides a progress bar callback to unpackers.
    """

    def on_start(self, count):
        """
        The unpack just started.
        """

        self.count = count
        self.current_file = ''

        widgets = [os.path.basename(self.unpacker.archive_path), ': ', SimpleProgress(), ' (', Percentage(), ')', CurrentFile(self, prefix=' - ')]
        self.progressbar = ProgressBar(widgets=widgets, maxval=count)
        self.progressbar.start()

    def on_update(self, current_file, progress):
        """
        The unpack was updated.
        """

        self.current_file = current_file
        self.progressbar.update(progress)

    def on_finish(self):
        """
        The unpack finished.
        """

        self.current_file = ''
        self.progressbar.update(self.count)
        self.progressbar.finish()

    def on_exception(self, exception):
        """
        The unpack failed.
        """

        if hasattr(self, 'progressbar'):
            self.on_finish()



"""
Teapot environment class.
"""

import os
import re
import sys
import json
import shlex
import hashlib

from contextlib import contextmanager

from teapot.log import LOGGER
from teapot.log import Highlight as hl


def perform_substitutions(value, parent_context):
    """
    Perform substitutions in the specified `value`.

    Values are taken from the `parent_context`. If no value is found in the
    parent context, the substitution is done with an empty string.

    Substitutions are done for $KEY on all platforms, and also for %KEY% on
    Windows.
    """

    if sys.platform.startswith('win32'):
        pattern = '\$(?P<unix_key>[A-Za-z_][A-Za-z_0-9]*)|%(?P<windows_key>[A-Za-z_][A-Za-z_0-9]*)%'
    else:
        pattern = '\$(?P<unix_key>[A-Za-z_][A-Za-z_0-9]*)'

    def substitute(match):
        key = match.group('unix_key') or match.group('windows_key')

        return parent_context.get(key, '')

    return re.sub(pattern, substitute, value)


class Environment(object):

    """
    Represents an environment for builders.
    """

    def __init__(self, name=None, variables=None, inherit=None, shell=None):
        """
        Initialize a new Environment.

        `name`, is the name of the environment.

        `variables`, if specified, is a map of the variables and their values.

        `inherit` is the environment to inherit from, or None, if an empty base
        environment is preferred.

        `shell`, if specified, is the shell to use to execute the commands
        during builds.

        `shell` may also be equal to True, in which case it will also inherit
        the shell from its inherited environment.
        """

        self.name = name or '<unnamed environment>'
        self.variables = variables or {}
        self.inherit = inherit
        self.shell = None

        if shell is True:
            if self.inherit:
                shell = self.inherit.shell
            else:
                shell = None

        if shell:
            if self.inherit:
                with self.inherit.enable(silent=True):
                    self.shell = map(lambda arg: perform_substitutions(arg, os.environ), shell)
            else:
                self.shell = map(lambda arg: perform_substitutions(arg, {}), shell)

    def __str__(self):
        """
        Get the name of the environment.
        """

        return self.name

    def __repr__(self):
        """
        Get a representation of the environment.
        """

        return '<%s.%s(name=%r, variables=%r, inherit=%r, shell=%r)>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.name,
            self.variables,
            self.inherit,
            self.shell,
        )

    @contextmanager
    def enable(self, silent=False):
        """
        Enable the environment and shell settings, if any.

        The current process environment is modified during the call, and
        restored when the function returns.

        If `silent` is truthy, no log output will be done.

        enable() is supposed to be used with a `with` statement.
        """

        saved_environ = os.environ.copy()

        try:
            if self.inherit:
                with self.inherit.enable(silent=True):
                    if not silent: LOGGER.info('Entering environment %s...', hl(self))

                    for key, value in self.variables.iteritems():
                        if value is not None:
                            os.environ[key] = perform_substitutions(value, saved_environ)
                        else:
                            if key in os.environ:
                                del os.environ[key]

                    yield self
            else:
                if not silent: LOGGER.info('Entering environment %s...', hl(self))

                os.environ.clear()

                for key, value in self.variables.iteritems():
                    if value is not None:
                        os.environ[key] = perform_substitutions(value, saved_environ)
                    else:
                        if key in saved_environ:
                            os.environ[key] = saved_environ.get(key)

                yield self

        finally:
            if not silent: LOGGER.info('Exiting environment %s...', hl(self))

            os.environ.clear()
            os.environ.update(saved_environ)

    @property
    def signature(self):
        """
        Get the signature of the environment.
        """

        data = {
            'name': self.name,
            'variables': self.variables,
            'inherit': self.inherit.signature if self.inherit else None,
            'shell': self.shell,
        }

        algorithm = hashlib.sha1()
        algorithm.update(json.dumps(data))

        result = algorithm.hexdigest()

        LOGGER.debug('%s\'s signature is: %s', hl(self), hl(result))

        return result



"""
Built-in extensions.
"""

import os
import sys

from teapot.path import windows_to_unix_path
from teapot.extensions.decorators import named_extension


@named_extension('root')
def root(builder, style='default'):
    """
    Get the builder prefix.
    """

    result = os.path.dirname(builder.attendee.party.path)

    if sys.platform.startswith('win32') and style == 'unix':
        result = windows_to_unix_path(result)

    return result

@named_extension('prefix')
def prefix(builder, style='default'):
    """
    Get the builder prefix.
    """

    result = os.path.join(
        builder.apply_extensions(builder.attendee.party.prefix),
        builder.apply_extensions(builder.attendee.prefix),
        builder.apply_extensions(builder.prefix),
    )

    if sys.platform.startswith('win32') and style == 'unix':
        result = windows_to_unix_path(result)

    return result

@named_extension('prefix_for')
def prefix_for(builder, for_attendee, for_builder='', style='default'):
    """
    Get the builder prefix for a given attendee, and optionally one of its builders.
    """

    party = builder.attendee.party

    for_attendee = party.get_attendee_by_name(for_attendee)
    builder_prefix = attendee.get_builder_by_name(for_builder).prefix if for_builder else ''

    result = os.path.join(
        builder.apply_extensions(party.prefix),
        builder.apply_extensions(attendee.prefix),
        builder.apply_extensions(builder_prefix),
    )

    if sys.platform.startswith('win32') and style == 'unix':
        result = windows_to_unix_path(result)

    return result

@named_extension('current_attendee')
def current_attendee(builder):
    """
    Get the current attendee.
    """

    return builder.attendee

@named_extension('current_builder')
def current_builder(builder):
    """
    Get the current builder.
    """

    return builder

@named_extension('current_archive_path')
def current_archive_path(builder, style='default'):
    """
    Get the current archive path.
    """

    result = builder.attendee.archive_path

    if sys.platform.startswith('win32') and style == 'unix':
        result = windows_to_unix_path(result)

    return result

@named_extension('current_source_tree_path')
def current_source_tree_path(builder, style='default'):
    """
    Get the current source tree path.
    """

    result = builder.attendee.source_tree_path

    if sys.platform.startswith('win32') and style == 'unix':
        result = windows_to_unix_path(result)

    return result



from setuptools import setup

import sys

install_requires = [
    'PyYaml',
    'requests>=1.2.3',
    'progressbar',
    'colorama',
    'LEPL',
]

if sys.version_info < (2, 7):
    install_requires.append('unittest2')

setup(
    name='teapot',
    url='http://teapot-builder.readthedocs.org/en/latest/index.html',
    author='Julien Kauffmann',
    author_email='julien.kauffmann@freelan.org',
    license='MIT License',
    version=open('VERSION').read().strip(),
    description="A multi-platform tool to automate the download, unpack and build of third-party softwares.",
    long_description="""\
teapot is a python library and a command-line tool that eases the
downloading, the unpacking and the building of third-party softwares in a
project.

teapot comes with the command-line tool `teapot` that does all the
necessary work.

teapot is designed with extension in mind, which means it is very easy
to add your own modules and extend its scope.
""",
    packages=[
        'teapot',
        'teapot.environments',
        'teapot.extensions',
        'teapot.fetchers',
        'teapot.unpackers',
        'teapot.builders',
        'teapot.filters',
        'teapot.defaults',
        'teapot.tests',
        'teapot.extra',
    ],
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'teapot = teapot.main:main',
        ],
    },
    test_suite='teapot.tests',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'License :: OSI Approved :: MIT License',
    ],
)



"""
An environment register class.
"""

from teapot.log import LOGGER
from teapot.log import Highlight as hl


class NoSuchEnvironmentError(ValueError):

    """
    The specified environment does not exist.
    """

    def __init__(self, name):
        """
        Create an NoSuchEnvironmentError for the specified `name`.
        """

        super(NoSuchEnvironmentError, self).__init__(
            'No such environment: %s' % name
        )

        self.name = name


class EnvironmentAlreadyRegisteredError(ValueError):

    """
    An environment was already registered with that name.
    """

    def __init__(self, name):
        """
        Create an EnvironmentAlreadyRegisteredError for the specified ` name`.
        """

        super(EnvironmentAlreadyRegisteredError, self).__init__(
            'An environment was already registered with that name: %s' % name
        )

        self.name = name


class EnvironmentRegister(object):

    """
    A environment register holds a reference to all existing environments and
    ensures no cyclic graph is created when they inherit one another.
    """

    def __init__(self):
        """
        Create a new register.
        """

        self.environments = {}

    def register_environment(self, name, environment):
        """
        Register the specified environment.

        `name` is the environment name to register as. It may differ from
        `environment.name`.

        `environment` is the environment to register.
        """

        LOGGER.debug('Registering environment "%s" as %s.', hl(environment.name), hl(name))

        if name in self.environments:
            raise EnvironmentAlreadyRegisteredError(name=name)

        self.environments[name] = environment

    def get_environment_by_name(self, name):
        """
        Get an environment by name, if it exists.

        If no environment has the specified name, a NoSuchEnvironmentError is raised.
        """

        if name in self.environments:
            return self.environments[name]

        raise NoSuchEnvironmentError(name=name)



"""
An unpacker that deals with .tgz files.
"""

from teapot.unpackers.base_unpacker import BaseUnpacker

import os
import tarfile

class InvalidTarballError(RuntimeError):

    """
    The specified archive is not a supported tarball.
    """

    def __init__(self, archive_path):
        """
        Create an InvalidTarballError for the specified `archive_path`.
        """

        super(InvalidTarballError, self).__init__(
            'Unsupported archive: %s' % archive_path
        )


class TarballHasNoCommonPrefixError(RuntimeError):

    """
    The specified archive does not contain a common prefix for its members.
    """

    def __init__(self, archive_path):
        """
        Create an TarballHasNoCommonPrefixError for the specified
        `archive_path`.
        """

        super(TarballHasNoCommonPrefixError, self).__init__(
            'The specified archive contains no common prefix: %s' % archive_path
        )


class TarballUnpacker(BaseUnpacker):

    """
    An unpacker class that deals with .tgz files.
    """

    types = [
        ('application/x-gzip', None),
        ('application/x-bzip2', None),
    ]

    def do_unpack(self):
        """
        Uncompress the archive.

        Return the path of the extracted folder.
        """

        if not tarfile.is_tarfile(self.archive_path):
            raise InvalidTarballError(archive_path=self.archive_path)

        tar = tarfile.open(self.archive_path, 'r')

        # We get the common prefix for all archive members.
        prefix = os.path.commonprefix(tar.getnames())

        # An archive member with the prefix as a name should exist in the archive.
        while True:
            try:
                prefix_member = tar.getmember(prefix)

                if prefix_member.isdir:
                    break

            except KeyError:
                pass

            new_prefix = os.path.dirname(prefix)

            if prefix == new_prefix:
                raise TarballHasNoCommonPrefixError(archive_path=self.archive_path)
            else:
                prefix = new_prefix

        source_tree_path = os.path.join(self.attendee.build_path, prefix_member.name)

        self.progress.on_start(count=len(tar.getmembers()))

        for index, member in enumerate(tar.getmembers()):
            if os.path.isabs(member.name):
                raise ValueError('Refusing to extract archive that contains absolute filenames.')

            self.progress.on_update(current_file=member.name, progress=index)
            tar.extract(member, path=self.attendee.build_path)

        self.progress.on_finish()

        return {
            'source_tree_path': source_tree_path,
        }



"""
teapot unit tests.
"""

import os
import sys

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from teapot.party import Party, NoSuchAttendeeError, CyclicDependencyError, load_party_file
from teapot.attendee import Attendee
from teapot.builders import Builder
from teapot.fetchers.callbacks import NullFetcherCallback
from teapot.unpackers.callbacks import NullUnpackerCallback
from teapot.environments import Environment, EnvironmentRegister, create_default_environment
from teapot.environments.environment_register import NoSuchEnvironmentError, EnvironmentAlreadyRegisteredError
from teapot.extensions import get_extension_by_name, parse_extension
from teapot.extensions.decorators import named_extension, ExtensionParsingError, DuplicateExtensionError, NoSuchExtensionError, ExtensionRecursionDetected


class TestTeapot(unittest.TestCase):

    def setUp(self):
        """
        Set up tests.
        """

        self.fixtures_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                'fixtures'
            )
        )

        self.party_file = os.path.join(self.fixtures_path, 'party.yaml')

    def test_parsing(self):
        """
        Test the parsing code.
        """

        party = load_party_file(self.party_file)

        party.fetcher_callback_class = NullFetcherCallback
        party.unpacker_callback_class = NullUnpackerCallback

        # Test the attendees.
        self.assertEqual(
            set([attendee.name for attendee in party.attendees]),
            set(['alpha', 'beta', 'gamma']),
        )

        alpha = party.get_attendee_by_name('alpha')
        beta = party.get_attendee_by_name('beta')
        gamma = party.get_attendee_by_name('gamma')

        self.assertEqual(len(alpha.sources), 1)
        self.assertEqual(len(alpha.builders), 3)
        self.assertEqual(len(alpha.enabled_builders), 2)

        self.assertEqual(len(beta.sources), 1)
        self.assertEqual(len(beta.builders), 1)
        self.assertEqual(len(beta.enabled_builders), 1)

        self.assertEqual(len(gamma.sources), 2)
        self.assertEqual(len(gamma.depends), 2)
        self.assertEqual(len(gamma.builders), 0)
        self.assertEqual(len(gamma.enabled_builders), 0)

        # Signature changes tests
        signatures = [builder.signature for builder in alpha.builders]

        # Changing filters should not change signature
        for builder in alpha.builders:
            builder.filters[:] = []

        for signature, builder in zip(signatures, alpha.builders):
            self.assertEqual(signature, builder.signature)

        # Changing commands must change the signature
        for builder in alpha.builders:
            builder.commands.append('foo')

        for signature, builder in zip(signatures, alpha.builders):
            self.assertNotEqual(signature, builder.signature)

        signatures = [builder.signature for builder in alpha.builders]

        # Changing prefix must change the signature
        for builder in alpha.builders:
            builder.prefix += 'foo'

        for signature, builder in zip(signatures, alpha.builders):
            self.assertNotEqual(signature, builder.signature)

        signatures = [builder.signature for builder in alpha.builders]

        # Changing environment must change the signature
        for builder in alpha.builders:
            builder.environment.variables['NEWVARIABLE'] = 'foo'

        for signature, builder in zip(signatures, alpha.builders):
            self.assertNotEqual(signature, builder.signature)

    def test_party(self):
        """
        The the party object.
        """

        def attendee_name(num):
            return 'attendee #%s' % num

        party = Party()

        # First test with a broken dependency graph.
        party.attendees = [
            Attendee(
                party=party,
                name=attendee_name(i),
                depends=[attendee_name(i + 1)],
            )
            for i in range(5)
        ]

        attendees_copy = list(party.attendees)

        with self.assertRaises(NoSuchAttendeeError) as context:
            party.get_ordered_attendees()

        self.assertEqual(context.exception.name, attendee_name(5))
        self.assertEqual(attendees_copy, party.attendees)

        # Test with a cyclic dependency graph.
        party.attendees.append(
            Attendee(
                party=party,
                name=attendee_name(5),
                depends=[attendee_name(2)],
            ),
        )

        attendees_copy = list(party.attendees)

        with self.assertRaises(CyclicDependencyError) as context:
            party.get_ordered_attendees()

        self.assertEqual(context.exception.cycle, party.attendees[2:])
        self.assertEqual(attendees_copy, party.attendees)

        # Test with a complete dependency graph.
        party.attendees[3].depends = []

        attendees_copy = list(party.attendees)

        self.assertEqual(
            party.get_ordered_attendees(),
            [party.attendees[i] for i in (3, 2, 1, 0, 5, 4)],
        )

        self.assertEqual(attendees_copy, party.attendees)

        self.assertEqual(
            party.get_ordered_attendees(attendees=[attendee_name(0)]),
            [party.attendees[i] for i in (3, 2, 1, 0)],
        )

        self.assertEqual(attendees_copy, party.attendees)

        self.assertEqual(
            party.get_ordered_attendees(attendees=[attendee_name(5), attendee_name(1)]),
            [party.attendees[i] for i in (3, 2, 1, 5)],
        )

        self.assertEqual(attendees_copy, party.attendees)


    def test_environments(self):
        """
        Test the environments.
        """

        os.environ['DUMMY'] = 'DUMMY1'
        os.environ['FOO'] = 'FOO1'
        os.environ['HELLO'] = 'HELLO1'

        default_environment = create_default_environment()

        # We test the variables before we enable the environment
        self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
        self.assertEqual(os.environ.get('FOO'), 'FOO1')
        self.assertEqual(os.environ.get('BAR'), None)
        self.assertEqual(os.environ.get('HELLO'), 'HELLO1')

        # We test that variables within the default environment are the same than before
        with default_environment.enable():
            self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
            self.assertEqual(os.environ.get('FOO'), 'FOO1')
            self.assertEqual(os.environ.get('BAR'), None)
            self.assertEqual(os.environ.get('HELLO'), 'HELLO1')

        self.assertEqual(default_environment.shell, None)

        environment = Environment(
            name='test_environment',
            variables={
                'FOO': 'FOO2',
                'BAR': 'BAR1',
                'HELLO': None,
                'FOO_EXTENDED': '$FOO-$FOO-$NON_EXISTING_VAR-$FOO_EXTENDED',
                'FOO_EXTENDED_PLATFORM': '[%HELLO%]',
                'NON_EXISTING': None,
            },
            inherit=default_environment,
            shell=['my shell'],
        )

        # We test the variables before we enable the environment
        self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
        self.assertEqual(os.environ.get('FOO'), 'FOO1')
        self.assertEqual(os.environ.get('BAR'), None)
        self.assertEqual(os.environ.get('HELLO'), 'HELLO1')
        self.assertEqual(os.environ.get('FOO_EXTENDED'), None)
        self.assertTrue('NON_EXISTING' not in os.environ)

        # We apply the environment and test those again
        with environment.enable():
            self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
            self.assertEqual(os.environ.get('FOO'), 'FOO2')
            self.assertEqual(os.environ.get('BAR'), 'BAR1')
            self.assertEqual(os.environ.get('HELLO'), None)
            self.assertEqual(os.environ.get('FOO_EXTENDED'), 'FOO1-FOO1--')
            self.assertTrue('NON_EXISTING' not in os.environ)

            if sys.platform.startswith('win32'):
                self.assertEqual(os.environ.get('FOO_EXTENDED_PLATFORM'), '[HELLO1]')
            else:
                self.assertEqual(os.environ.get('FOO_EXTENDED_PLATFORM'), '[%HELLO%]')

        self.assertEqual(environment.shell, ['my shell'])

        sub_environment = Environment(
            name='test_environment',
            inherit=environment,
            shell=True,
        )

        self.assertEqual(sub_environment.shell, ['my shell'])

        orphan_environment = Environment(
            name='test_environment',
            variables={
                'FOO': 'FOO3',
                'BAR': 'BAR1',
                'HELLO': None,
                'NON_EXISTING': None,
            },
            inherit=None,
        )

        # We test the variables before we enable the environment
        self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
        self.assertEqual(os.environ.get('FOO'), 'FOO1')
        self.assertEqual(os.environ.get('BAR'), None)
        self.assertEqual(os.environ.get('HELLO'), 'HELLO1')
        self.assertTrue('NON_EXISTING' not in os.environ)

        # We apply the environment and test those again
        with orphan_environment.enable():
            self.assertEqual(os.environ.get('DUMMY'), None)
            self.assertEqual(os.environ.get('FOO'), 'FOO3')
            self.assertEqual(os.environ.get('BAR'), 'BAR1')
            self.assertEqual(os.environ.get('HELLO'), 'HELLO1')
            self.assertTrue('NON_EXISTING' not in os.environ)

        self.assertEqual(orphan_environment.shell, None)

        shell_environment = Environment(
            name='test_environment',
            variables={
                'FOO': 'FOO3',
            },
            inherit=default_environment,
            shell=['$FOO'],
        )

        self.assertEqual(shell_environment.shell, ['FOO1'])

        orphan_shell_environment = Environment(
            name='test_environment',
            variables={
                'FOO': 'FOO3',
            },
            shell=['$FOO'],
        )

        self.assertEqual(orphan_shell_environment.shell, [''])

        # Test signatures changes

        signature = environment.signature
        sub_signature = sub_environment.signature

        environment.variables['NEWVARIABLE'] = 'bar'

        self.assertNotEqual(signature, environment.signature)
        self.assertNotEqual(sub_environment, sub_environment.signature)

        signature = environment.signature
        sub_signature = sub_environment.signature

        environment.shell = ['foo', 'goo']

        self.assertNotEqual(signature, environment.signature)
        self.assertNotEqual(sub_environment, sub_environment.signature)

        signature = environment.signature
        sub_signature = sub_environment.signature

        sub_environment.variables['NEWVARIABLE'] = 'bar'

        self.assertEqual(signature, environment.signature)
        self.assertNotEqual(sub_environment, sub_environment.signature)

        signature = environment.signature
        sub_signature = sub_environment.signature

        sub_environment.shell = ['foo', 'goo', 'boo']

        self.assertEqual(signature, environment.signature)
        self.assertNotEqual(sub_environment, sub_environment.signature)

    def test_environment_register(self):
        """
        Test the environment register.
        """

        register = EnvironmentRegister()

        with self.assertRaises(NoSuchEnvironmentError) as context:
            register.get_environment_by_name('foo')

        self.assertEqual(context.exception.name, 'foo')

        environments = []

        for index in xrange(50):
            environment = Environment(
                name='test_environment_%s' % index,
            )

            environments.append(environment)
            register.register_environment('env%s' % index, environment)

        with self.assertRaises(EnvironmentAlreadyRegisteredError) as context:
            register.register_environment('env5', environment)

        self.assertEqual(context.exception.name, 'env5')

        for index, environment in enumerate(environments):
            self.assertEqual(register.get_environment_by_name('env%s' % index), environment)

    def test_extensions(self):
        """
        Test the extensions.
        """

        extension_name = 'test_extension'
        result_dict = {
            'foo': None,
            'bar': None,
            'builder': None,
        }

        @named_extension(extension_name)
        def test_extension(builder, foo, bar):
            result_dict['builder'] = builder
            result_dict['foo'] = foo
            result_dict['bar'] = bar

        extension = get_extension_by_name(extension_name)

        # We make sure the extension returned is the one we just defined.
        self.assertEqual(extension, test_extension)

        # We make sure the extension parsing code works.
        parse_extension('%s(one, two)' % extension_name, builder='three')

        self.assertEqual(result_dict['foo'], 'one')
        self.assertEqual(result_dict['bar'], 'two')
        self.assertEqual(result_dict['builder'], 'three')

        # Parsing an extension with less than the expected count of parameters raises a TypeError.
        self.assertRaises(TypeError, parse_extension, '%s(one)' % extension_name, builder='two')

        # Parsing an extension with more than the expected count of parameters raises a TypeError.
        self.assertRaises(TypeError, parse_extension, '%s(one, two, three)' % extension_name, builder='four')

        # Parsing an unbalanced extension raises an ExtensionParsingError.
        self.assertRaises(ExtensionParsingError, parse_extension, '%s(one, two' % extension_name, builder='three')

        # Requesting a non-existing extension raises a NoSuchExtensionError.
        self.assertRaises(NoSuchExtensionError, parse_extension, '%s(one, two)' % 'non_existing_extension_name', builder='three')
        self.assertRaises(NoSuchExtensionError, get_extension_by_name, 'non_existing_extension_name')

        # Registering an already existing extension raises a DuplicateExtensionError.
        self.assertRaises(DuplicateExtensionError, named_extension, extension_name)

        # Registering an already existing extension with override does not throw.
        @named_extension(extension_name, override=True)
        def test_extension(builder, foo=None, bar=None):
            result_dict['builder'] = builder
            result_dict['foo'] = foo
            result_dict['bar'] = bar

        # We make sure the extension parsing code works even for function that can take no arguments.
        parse_extension('%s()' % extension_name, builder='one')

        self.assertEqual(result_dict['foo'], None)
        self.assertEqual(result_dict['bar'], None)
        self.assertEqual(result_dict['builder'], 'one')

        # We make sure the extension parsing code works even for function that can take no arguments, when parenthesis are not specified.
        parse_extension(extension_name, builder='two')

        self.assertEqual(result_dict['foo'], None)
        self.assertEqual(result_dict['bar'], None)
        self.assertEqual(result_dict['builder'], 'two')

        # We test extension usage in builders.
        builder = Builder(
            attendee=None,
            name='test_builder',
            tags=[],
            commands=['dummy'],
            environment=create_default_environment(),
            clean_commands=['calling {{non_existing_extension_name}} somehow'],
        )

        self.assertRaises(NoSuchExtensionError, builder.clean)

        # We test extension recursivity failsafes.
        @named_extension(extension_name + '_one')
        def test_extension_one(builder):
            return builder.apply_extensions('foo {{%s_two}} bar' % extension_name)

        @named_extension(extension_name + '_two')
        def test_extension_two(builder):
            return builder.apply_extensions('foo2 {{%s_one}} bar2' % extension_name)

        builder = Builder(
            attendee=None,
            name='test_builder',
            tags=[],
            commands=['dummy'],
            environment=create_default_environment(),
            clean_commands=['calling {{%s_one}} somehow' % extension_name],
        )

        self.assertRaises(ExtensionRecursionDetected, builder.clean)

        # We test extension recursive replacement.
        @named_extension(extension_name + '_two', override=True)
        def test_extension_two(builder):
            return 'two_called'

        command = builder.apply_extensions(builder.clean_commands[0])

        self.assertEqual(command, 'calling foo two_called bar somehow')


if __name__ == '__main__':
    unittest.main()



"""
The extensions decorators.
"""

import re
import itertools

from functools import wraps


EXTENSIONS = {}

class NoSuchExtensionError(ValueError):

    """
    No extension exists with the specified name.
    """

    def __init__(self, name):
        """
        Create an NoSuchExtensionError for the specified extension `name`.
        """

        self.name = name

        super(NoSuchExtensionError, self).__init__(
            'No extension with the specified name: %s' % name
            )


class DuplicateExtensionError(ValueError):

    """
    Another extension with the same name exists.
    """

    def __init__(self, name):
        """
        Create a DuplicateExtensionError for the specified extension `name`.
        """

        self.name = name

        super(DuplicateExtensionError, self).__init__(
            'Another extension was already registered with the name %r' % name
            )

class ExtensionParsingError(ValueError):

    """
    No valid extension could be parsed.
    """

    def __init__(self, code):
        """
        Create an ExtensionParsingError for the specified `code`.
        """

        self.code = code

        super(ExtensionParsingError, self).__init__(
            'No extension could be parsed from %r' % code,
        )


class ExtensionRecursionDetected(ValueError):

    """
    A recursion was detected in the extension callstack.
    """

    def __init__(self, callstack):
        """
        Create an ExtensionRecursionDetected for the specified `callstack`.
        """

        self.callstack = callstack

        callstack_output = '\n'.join(
            '%s(%s)' % (
                stack[0],
                ', '.join(map(repr, stack[1:])),
            )
            for stack
            in self.callstack
        )

        super(ExtensionRecursionDetected, self).__init__(
            'Extension recursion detected. Callstack was:\n' + callstack_output,
        )


class named_extension(object):

    """
    Registers a function to be a extension.
    """

    CALLSTACK = []

    def __init__(self, name, override=False, infinite_recursion_allowed=False):
        """
        Registers the function with the specified name.

        If another function was registered with the same name, a
        `DuplicateExtensionError` will be raised, unless `override` is truthy.
        """

        if name in EXTENSIONS and not override:
            raise DuplicateExtensionError(name)

        self.name = name
        self.infinite_recursion_allowed = infinite_recursion_allowed

    def __call__(self, func):
        """
        Registers the function and returns it unchanged.
        """

        @wraps(func)
        def decorated(*args):

            call_signature = tuple(itertools.chain([self.name], args))

            if not self.infinite_recursion_allowed and call_signature in self.CALLSTACK:
                raise ExtensionRecursionDetected(self.CALLSTACK + [call_signature])

            self.CALLSTACK.append(call_signature)

            try:
                return func(*args)

            finally:
                self.CALLSTACK.pop()

        EXTENSIONS[self.name] = decorated

        return decorated

def get_extension_by_name(name):
    """
    Get a extension by name.

    If no extension matches the specified name, an NoSuchExtensionError is raised.
    """

    if not name in EXTENSIONS:
        raise NoSuchExtensionError(name=name)

    return EXTENSIONS[name]

def parse_extension(code, builder):
    """
    Parse an extension call in `code` and calls it.

    `builder` will be passed to the extension call as the `builder` argument.

    The function returns the result of the extension call.
    """

    match = re.match(r'^(?P<name>[a-zA-Z_]+)(\((?P<args>(|(\s*\w+\s*)(,\s*\w+\s*)*))\)|)$', code)

    if not match:
        raise ExtensionParsingError(code=code)

    name = match.group('name')
    args = match.group('args')

    if args:
        args = [arg.strip() for arg in args.split(',')]
    else:
        args = []

    extension = get_extension_by_name(name)

    return extension(builder, *args)



"""
The defaults.
"""

import sys

if sys.platform in ('win32', 'cygwin'):
    from teapot.defaults.windows import *
elif sys.platform in ('darwin'):
    from teapot.defaults.osx import *
elif sys.platform.startswith('linux'):
    from teapot.defaults.linux import *
